# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  swagger.py
@Time    :  2022/7/18 10:16
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import copy
import json
import re
import urllib.parse
import uuid


class Swagger2:
    def __init__(self, source, deep=5):
        self.source = source
        self.deep = deep

        self.__scheme = self.schemes[0]
        self.__host = self.source.get("host") or "localhost"

    def __property__(self, prop):
        if isinstance(prop, dict):
            _type = prop.get("type")
            _format = prop.get("format")
            _example = prop.get("example")
            if _example:
                prop = _example
            elif _type == "integer":
                prop = 0
            elif _type == "number":
                prop = 1.75
            elif _type == "boolean":
                prop = True
            elif _type == "string":
                if _format in [None, "byte", "binary", "password"]:
                    prop = "string"
                elif _format == "date":
                    prop = "1970-01-01"
                elif _format == "date-time":
                    prop = "1970-01-01 00:00:00"
            elif _type == "file":
                prop = "file.txt"
            elif _type == "array":
                prop = []

        return prop

    @property
    def schemes(self):
        return self.source.get("schemes") or ["http"]

    @property
    def scheme(self):
        return self.__scheme

    @scheme.setter
    def scheme(self, value):
        self.__scheme = value

    @property
    def host(self):
        return self.__host

    @host.setter
    def host(self, value):
        self.__host = value

    @property
    def base_path(self):
        return self.source.get("base_path") or "/"

    @property
    def models(self):
        ref_results = {}

        definitions = copy.deepcopy(self.source.get("definitions"))
        for _ in range(self.deep):
            for key in list(definitions.keys()):
                properties = definitions[key].get("properties")
                if properties:
                    ref_keys = re.findall(r'"#/definitions/(.+?)"', json.dumps(properties, ensure_ascii=False))
                    if ref_keys and len(ref_keys) > 0:
                        if set(ref_results.keys()).issuperset(set(ref_keys)):
                            for prop_name, prop in properties.items():
                                ref_key = prop.get("$ref")
                                if ref_key:
                                    properties[prop_name] = ref_results.get(ref_key.split("/").pop())
                                else:
                                    properties[prop_name] = self.__property__(prop)
                            ref_results.update({key: properties})
                            del definitions[key]
                    else:
                        for prop_name, prop in properties.items():
                            properties[prop_name] = self.__property__(prop)
                        ref_results.update({key: properties})
                        del definitions[key]
                else:
                    ref_results.update({key: {}})
                    del definitions[key]
            # 无模型停止递归
            if len(definitions) == 0:
                break
        # if len(definitions) > 0:
        #     raise IncompleteModelError('The model is incomplete. Please adjust the deep value of the swagger object.')
        return ref_results

    @property
    def apis(self):
        models = self.models

        api_results = []

        for path, forms in self.source.get("paths").items():
            for method, form in forms.items():
                # 接口名称
                name = form.get("summary")
                # 内容类型
                content_type = form.get("consumes") or [""]
                content_type = content_type[0]
                # 接口路径
                _path = (self.base_path + path).replace("//", "/")
                # 拼接后的URL
                url = urllib.parse.urlunparse((self.scheme, self.host, _path, None, None, None))
                _api = {
                    "id": uuid.uuid4().hex,
                    "name": name,
                    "method": method,
                    "path": _path,
                    "url": url,
                    "headers": {"Content-Type": content_type},
                    "paths": {},
                    "query": {},
                    "json": {},
                    "form": {},
                    "formData": {},
                }
                # 参数提取
                parameters = form.get("parameters")
                if parameters:
                    for param in parameters:
                        _in = param.get("in")
                        param_name = param.get("name")
                        param_value = self.__property__(param)

                        if isinstance(param_value, dict):
                            schema = param_value.get("schema")
                            if schema and schema.get("type") == "array":
                                param_value = []
                            else:
                                ref_key = re.search(r'"#/definitions/(.+?)"', json.dumps(schema, ensure_ascii=False))
                                if ref_key:
                                    param_value = models.get(ref_key.group(1))

                        if _in == "header":
                            _api.get("headers").update({param_name: param_value})
                        elif _in == "path":
                            _api.get("paths").update({param_name: param_value})
                        elif _in == "query":
                            _api.get("query").update({param_name: param_value})
                        elif _in == "body":
                            _api.update({"json": param_value})
                        elif _in == "formData":
                            if "multipart/form-data" in content_type:
                                _api.get("formData").update({param_name: param_value})
                            else:
                                _api.get("form").update({param_name: param_value})

                api_results.append(_api)
        return api_results



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  stoppable_thread.py
@Time    :  2022/5/15 10:37 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import contextlib
import ctypes
import os
import threading

__all__ = ("StoppableThread", "JoinThread")


class StoppableThread(threading.Thread):
    def stop_thread(self, exception, repeat_every=2.0):
        if self.is_alive() is False:
            return True

        self._stderr = open(os.devnull, "w")

        join_thread = JoinThread(self, exception, repeat_every=repeat_every)
        join_thread._stderr = self._stderr
        join_thread.start()
        join_thread._stderr = self._stderr
        return None

    def stop(self, exception, repeat_every=2.0):
        return self.stop_thread(exception, repeat_every)


class JoinThread(threading.Thread):
    def __init__(self, other_thread, exception, repeat_every=2.0):
        threading.Thread.__init__(self)
        self.other_thread = other_thread
        self.exception = exception
        self.repeat_every = repeat_every
        self.daemon = True

    def run(self):
        self.other_thread._Thread__stderr = self._stderr
        if hasattr(self.other_thread, "_thread__stop"):
            self.other_thread._thread__stop()
        while self.other_thread.is_alive():
            ctypes.pythonapi.PyThreadState_SetAsyncExc(
                ctypes.c_long(self.other_thread.ident),
                ctypes.py_object(self.exception),
            )
            self.other_thread.join(self.repeat_every)

        with contextlib.suppress(Exception):
            self._stderr.close()



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  bases.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from abc import ABC, abstractmethod
from collections import ChainMap
from dataclasses import dataclass
from functools import wraps
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Mapping,
    Sequence,
    Type,
    TypeVar,
    cast,
)

from pydantic import BaseModel, create_model
from pydantic.generics import GenericModel
from pydantic.types import conint

T = TypeVar("T")
C = TypeVar("C")

TAbstractPage = TypeVar("TAbstractPage", bound="AbstractPage")


@dataclass
class RawParams:
    limit: int
    offset: int


class AbstractParams(ABC):
    @abstractmethod
    def to_raw_params(self) -> RawParams:
        pass


def _create_params(cls: Type[AbstractParams], fields: Dict[str, Any]) -> Mapping[str, Any]:
    if not issubclass(cls, BaseModel):
        raise ValueError(f"{cls.__name__} must be subclass of BaseModel")

    incorrect = sorted(fields.keys() - cls.__fields__.keys())
    if incorrect:
        ending = "s" if len(incorrect) > 1 else ""
        raise ValueError(f"Unknown field{ending} {', '.join(incorrect)}")

    anns = ChainMap(*(obj.__dict__.get("__annotations__", {}) for obj in cls.mro()))
    return {name: (anns[name], val) for name, val in fields.items()}


class AbstractPage(GenericModel, Generic[T], ABC):
    __params_type__: ClassVar[Type[AbstractParams]]

    @classmethod
    @abstractmethod
    def create(cls: Type[C], items: Sequence[T], total: int, params: AbstractParams) -> C:
        pass

    @classmethod
    def with_custom_options(cls: Type[TAbstractPage], **kwargs: Any) -> Type[TAbstractPage]:
        params_cls = cast(Type[AbstractPage], cls).__params_type__

        custom_params: Any = create_model(
            params_cls.__name__,
            __base__=params_cls,
            **_create_params(params_cls, kwargs),
        )

        if cls.__concrete__:
            bases = (cls,)
        else:
            params = tuple(TypeVar(f"T{i}") for i, _ in enumerate(cls.__parameters__))
            bases = (cls[params], Generic[params])  # type: ignore

        @wraps(cls, updated=())
        class CustomPage(*bases):  # type: ignore
            __params_type__: ClassVar[Type[AbstractParams]] = custom_params

        return cast(Type[TAbstractPage], CustomPage)

    class Config:
        arbitrary_types_allowed = True


class BasePage(AbstractPage[T], Generic[T], ABC):
    items: Sequence[T]
    total: conint(ge=0)  # type: ignore


__all__ = [
    "AbstractPage",
    "AbstractParams",
    "BasePage",
    "RawParams",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  sync_sqlalchemy.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from __future__ import annotations

from typing import Any, Optional, TypeVar

from sqlalchemy.orm import Query
from sqlalchemy.sql import Select

from .api import create_page, resolve_params
from .bases import AbstractPage, AbstractParams

T = TypeVar("T", Select, Query)


def paginate_query(query: T, params: AbstractParams) -> T:
    raw_params = params.to_raw_params()
    return query.limit(raw_params.limit).offset(raw_params.offset)


def _to_dict(obj: Any) -> Any:
    try:
        return obj._asdict()
    except AttributeError:
        return obj


def paginate(query: Query, params: Optional[AbstractParams] = None) -> AbstractPage:
    params = resolve_params(params)

    total = query.count()
    items = [_to_dict(item) for item in paginate_query(query, params)]

    return create_page(items, total, params)


__all__ = ["paginate_query", "paginate"]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/20 2:52 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from .hitfilter import DFAFilter

__all__ = ["DFAFilter"]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  enums.py
@Time    :  2022/5/2 2:52 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""


class GlobalVarEnum:
    APP_NAME = "custard"
    BEYOND_RESTRICT_MSG = "The interview is too fast, please have a cup of tea and take a break!"
    INIT_ERR_MSG = "You must call Limiter.init in startup event of fastapi!"



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  moment.py
@Time    :  2020/9/17 19:05
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  日期处理类库
"""

import datetime
import time


class Moment:
    @staticmethod
    def get_now_time(layout="%Y-%m-%d %H:%M:%S") -> str:
        """
        获取当前时间
        Args:
            layout: 10timestamp, 13timestamp,  else  时间类型
        Returns:
        Examples:
            >>> print(Moment.get_now_time("%Y-%m-%d %H:%M:%S"))
            >>> print(Moment.get_now_time("10timestamp"))
            >>> print(Moment.get_now_time("13timestamp"))
        """
        tim = datetime.datetime.now()
        temp = tim.strftime("%Y-%m-%d %H:%M:%S")
        # 获取10位时间戳
        if layout == "10timestamp":
            tim = int(time.mktime(time.strptime(temp, "%Y-%m-%d %H:%M:%S")))
        # 获取13位时间戳
        elif layout == "13timestamp":
            datetime_object = datetime.datetime.now()
            now_timetuple = datetime_object.timetuple()
            now_second = time.mktime(now_timetuple)
            tim = int(now_second * 1000 + datetime_object.microsecond / 1000)
        # 按传入格式获取时间
        else:
            tim = tim.strftime(layout)
        return tim

    @staticmethod
    def time_format(custom, time_format_str="%Y-%m-%d %H:%M:%S"):
        """
        降级格式化时间
        Args:
            custom:
            time_format_str:
        Returns:
        Examples:
            >>> print(Moment.time_format(custom="2022-03-30 11:05:2"))
        """
        try:
            today = datetime.datetime.strptime(custom, time_format_str)
            # 以下异常仅限于兼容目的 、开发返回的值不规范时使用、层层降级(若是还有问题就是真的格式有问题)
        except ValueError:
            blank_split = custom.split(" ")
            format_str = time_format_str[:-2] if len(blank_split) == 2 and len(custom) <= 13 else time_format_str[:-3]
            today = Moment.time_format(custom=custom, time_format_str=format_str)
        return today

    @staticmethod
    def skew_date(
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
        custom=None,
        time_format_str="%Y-%m-%d %H:%M:%S",
    ):
        """
        日期偏移
        Args:
            days:
            seconds:
            microseconds:
            milliseconds:
            minutes:
            hours:
            weeks:
            custom:
            time_format_str:
        Returns:
        Examples:
            >>> import random
            >>> custom_time = ["2022-03-1 11:15:15","2022-03-1 11:15",
            ... "2022-03-1 11","2022-03-1 ","2022-03-1"]
            >>> for index in custom_time:
            ...     Moment.skew_date(hours=random.randint(1,5), custom=index)
        """
        if custom is not None:
            today = Moment.time_format(custom=custom, time_format_str=time_format_str)
        else:
            today = datetime.datetime.now()
        return (today + datetime.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)).strftime(
            time_format_str,
        )

    @staticmethod
    def timestamp_to_date(timestamp):
        """
        时间戳格式化为xxx年xx月xx日
        Args:
            timestamp:
        Returns:
        Examples:
            >>> print(Moment.timestamp_to_date(1603282677.5209892))
        """
        if not isinstance(timestamp, (int, float)):
            return None
        time_tuple = time.localtime(timestamp)
        specific_data = (
            str(time_tuple[0])
            + "-"
            + str(time_tuple[1])
            + "-"
            + str(time_tuple[2])
            + " "
            + str(time_tuple[3])
            + ":"
            + str(time_tuple[4])
            + ":"
            + str(time_tuple[5])
        )
        return specific_data

    @staticmethod
    def calc_time_diff(start, end):
        """
        中间时差时段计算
        Args:
            start: 开始日期
            end:   结束日期
        Returns:
        Examples:
            >>> print(Moment.calc_time_diff("2020-06-05", "2020-07-01"))
        """
        date_list = []
        begin_date = datetime.datetime.strptime(start, "%Y-%m-%d")
        end_date = datetime.datetime.strptime(end, "%Y-%m-%d")
        while begin_date <= end_date:
            date_str = begin_date.strftime("%Y-%m-%d")
            date_list.append(date_str)
            begin_date += datetime.timedelta(days=1)
        return date_list

    @staticmethod
    def time_to_timestamp(format_time):
        """
        年月日时分秒转化为时间戳
        Args:
            format_time:
        Returns:
        Examples:
            >>> print(Moment.time_to_timestamp("2020-06-01 18:50:00"))
        """
        return int(time.mktime(time.strptime(format_time, "%Y-%m-%d %H:%M:%S")))

    @staticmethod
    def set_sleep_time(timestamp):
        """
        休眠xx时间
        Args:
            timestamp:
        Returns:
        """
        time.sleep(timestamp)

    @staticmethod
    def compare_time(time1, time2):
        """
        时间比较
        Args:
            time1:
            time2:
        Returns:
        """ """
        Examples:
            >>> print(Moment.compare_time("2021-08-23 17:11:37", "2021-08-22 17:11:37"))
        """
        time1 = datetime.datetime.strptime(time1, "%Y-%m-%d %H:%M:%S")
        time2 = datetime.datetime.strptime(time2, "%Y-%m-%d %H:%M:%S")
        return time1 > time2



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  jenkins.py
@Time    :  2022/5/30 1:35 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  Jenkins工具驱动
"""

import contextlib
import json
import logging
import operator
import os
import re
import socket
import sys
import time
import warnings
import xml.etree.ElementTree as ET
from urllib import parse
from urllib.parse import quote

import pkg_resources
import requests
import requests.exceptions as req_exc

from custard.utils.multi import MultiKeyDict

try:
    import requests_kerberos
except ImportError:
    requests_kerberos = None

# Set default logging handler to avoid "No handler found" warnings.
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())

if sys.version_info < (2, 7, 0):
    warnings.warn("Support for python 2.6 is deprecated and will be removed.")

LAUNCHER_SSH = "hudson.plugins.sshslaves.SSHLauncher"
LAUNCHER_COMMAND = "hudson.slaves.CommandLauncher"
LAUNCHER_JNLP = "hudson.slaves.JNLPLauncher"
LAUNCHER_WINDOWS_SERVICE = "hudson.os.windows.ManagedWindowsServiceLauncher"
DEFAULT_HEADERS = {"Content-Type": "text/xml; charset=utf-8"}

# REST Endpoints
INFO = "api/json"
PLUGIN_INFO = "pluginManager/api/json?depth=%(depth)s"
CRUMB_URL = "crumbIssuer/api/json"
WHOAMI_URL = "me/api/json?depth=%(depth)s"
JOBS_QUERY = "?tree=%s"
JOBS_QUERY_TREE = "jobs[url,color,name,%s]"
JOB_INFO = "%(folder_url)sjob/%(short_name)s/api/json?depth=%(depth)s"
JOB_NAME = "%(folder_url)sjob/%(short_name)s/api/json?tree=name"
ALL_BUILDS = "%(folder_url)sjob/%(short_name)s/api/json?tree=allBuilds[number,url]"
Q_INFO = "queue/api/json?depth=0"
Q_ITEM = "queue/item/%(number)d/api/json?depth=%(depth)s"
CANCEL_QUEUE = "queue/cancelItem?id=%(id)s"
CREATE_JOB = "%(folder_url)screateItem?name=%(short_name)s"  # also post config.xml
CONFIG_JOB = "%(folder_url)sjob/%(short_name)s/config.xml"
DELETE_JOB = "%(folder_url)sjob/%(short_name)s/doDelete"
ENABLE_JOB = "%(folder_url)sjob/%(short_name)s/enable"
DISABLE_JOB = "%(folder_url)sjob/%(short_name)s/disable"
CHECK_JENKINSFILE_SYNTAX = "pipeline-model-converter/validateJenkinsfile"
SET_JOB_BUILD_NUMBER = "%(folder_url)sjob/%(short_name)s/nextbuildnumber/submit"
COPY_JOB = "%(from_folder_url)screateItem?name=%(to_short_name)s&mode=copy&from=%(from_short_name)s"
RENAME_JOB = "%(from_folder_url)sjob/%(from_short_name)s/doRename?newName=%(to_short_name)s"
BUILD_JOB = "%(folder_url)sjob/%(short_name)s/build"
STOP_BUILD = "%(folder_url)sjob/%(short_name)s/%(number)s/stop"
BUILD_WITH_PARAMS_JOB = "%(folder_url)sjob/%(short_name)s/buildWithParameters"
BUILD_INFO = "%(folder_url)sjob/%(short_name)s/%(number)d/api/json?depth=%(depth)s"
BUILD_CONSOLE_OUTPUT = "%(folder_url)sjob/%(short_name)s/%(number)d/consoleText"
BUILD_ENV_VARS = "%(folder_url)sjob/%(short_name)s/%(number)d/injectedEnvVars/api/json" + "?depth=%(depth)s"
BUILD_TEST_REPORT = "%(folder_url)sjob/%(short_name)s/%(number)d/testReport/api/json" + "?depth=%(depth)s"
DELETE_BUILD = "%(folder_url)sjob/%(short_name)s/%(number)s/doDelete"
WIPEOUT_JOB_WORKSPACE = "%(folder_url)sjob/%(short_name)s/doWipeOutWorkspace"
NODE_LIST = "computer/api/json?depth=%(depth)s"
CREATE_NODE = "computer/doCreateItem"
DELETE_NODE = "computer/%(name)s/doDelete"
NODE_INFO = "computer/%(name)s/api/json?depth=%(depth)s"
NODE_TYPE = "hudson.slaves.DumbSlave$DescriptorImpl"
TOGGLE_OFFLINE = "computer/%(name)s/toggleOffline?offlineMessage=%(msg)s"
CONFIG_NODE = "computer/%(name)s/config.xml"
VIEW_NAME = "%(folder_url)sview/%(short_name)s/api/json?tree=name"
VIEW_JOBS = "view/%(name)s/api/json?tree=jobs[url,color,name]"
CREATE_VIEW = "%(folder_url)screateView?name=%(short_name)s"
CONFIG_VIEW = "%(folder_url)sview/%(short_name)s/config.xml"
DELETE_VIEW = "%(folder_url)sview/%(short_name)s/doDelete"
SCRIPT_TEXT = "scriptText"
NODE_SCRIPT_TEXT = "computer/%(node)s/scriptText"
PROMOTION_NAME = "%(folder_url)sjob/%(short_name)s/promotion/process/%(name)s/api/json?tree=name"
PROMOTION_INFO = "%(folder_url)sjob/%(short_name)s/promotion/api/json?depth=%(depth)s"
DELETE_PROMOTION = "%(folder_url)sjob/%(short_name)s/promotion/process/%(name)s/doDelete"
CREATE_PROMOTION = "%(folder_url)sjob/%(short_name)s/promotion/createProcess?name=%(name)s"
CONFIG_PROMOTION = "%(folder_url)sjob/%(short_name)s/promotion/process/%(name)s/config.xml"
LIST_CREDENTIALS = (
    "%(folder_url)sjob/%(short_name)s/credentials/store/folder/" "domain/%(domain_name)s/api/json?tree=credentials[id]"
)
CREATE_CREDENTIAL = (
    "%(folder_url)sjob/%(short_name)s/credentials/store/folder/" "domain/%(domain_name)s/createCredentials"
)
CONFIG_CREDENTIAL = (
    "%(folder_url)sjob/%(short_name)s/credentials/store/folder/" "domain/%(domain_name)s/credential/%(name)s/config.xml"
)
CREDENTIAL_INFO = (
    "%(folder_url)sjob/%(short_name)s/credentials/store/folder/"
    "domain/%(domain_name)s/credential/%(name)s/api/json?depth=0"
)
QUIET_DOWN = "quietDown"


class JenkinsException(Exception):
    """General exception type for jenkins-API-related failures."""


class NotFoundException(JenkinsException):
    """A special exception to call out the case of receiving a 404."""


class EmptyResponseException(JenkinsException):
    """A special exception to call out the case receiving an empty response."""


class BadHTTPException(JenkinsException):
    """A special exception to call out the case of a broken HTTP response."""


class TimeoutException(JenkinsException):
    """A special exception to call out in the case of a socket timeout."""


class WrappedSession(requests.Session):
    """A wrapper for requests.Session to override 'verify' property, ignoring REQUESTS_CA_BUNDLE environment variable."""

    def mergeEnvSettings(self, url, proxies, stream, verify, *args, **kwargs):
        if self.verify is False:
            verify = False

        return super(WrappedSession, self).merge_environment_settings(url, proxies, stream, verify, *args, **kwargs)


class Plugin(dict):
    """Dictionary object containing plugin metadata."""

    def __init__(self, *args, **kwargs):
        """Populates dictionary using json object input.

        accepts same arguments as python `dict` class.
        """
        version = kwargs.pop("version", None)

        super(Plugin, self).__init__(*args, **kwargs)
        self["version"] = version

    def __setitem__(self, key, value):
        """Overrides default setter to ensure that the version key is always
        a PluginVersion class to abstract and simplify version comparisons
        """
        if key == "version":
            value = PluginVersion(value)
        super(Plugin, self).__setitem__(key, value)


class PluginVersion(str):
    """Class providing comparison capabilities for plugin versions."""

    _VERSION_RE = re.compile(r"(.*)-(?:SNAPSHOT|BETA)")

    def __init__(self, version):
        """Parse plugin version and store it for comparison."""

        self._version = version
        self.parsed_version = pkg_resources.parse_version(self.__convert_version(version))

    def __convert_version(self, version):
        return self._VERSION_RE.sub(r"\g<1>.preview", str(version))

    def __compare(self, op, version):
        return op(
            self.parsed_version,
            pkg_resources.parse_version(self.__convert_version(version)),
        )

    def __le__(self, version):
        return self.__compare(operator.le, version)

    def __lt__(self, version):
        return self.__compare(operator.lt, version)

    def __ge__(self, version):
        return self.__compare(operator.ge, version)

    def __gt__(self, version):
        return self.__compare(operator.gt, version)

    def __eq__(self, version):
        return self.__compare(operator.eq, version)

    def __ne__(self, version):
        return self.__compare(operator.ne, version)

    def __str__(self):
        return str(self._version)

    def __repr__(self):
        return str(self._version)


class Jenkins(object):
    timeoutWarningIssued = False

    def __init__(self, url, username=None, password=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        """Create handle to Jenkins instance.

        All methods will raise :class:`JenkinsException` on failure.

        :param url: URL of Jenkins server, ``str``
        :param username: Server username, ``str``
        :param password: Server password, ``str``
        :param timeout: Server connection timeout in secs (default: not set), ``int``
        """
        if url[-1] == "/":
            self.server = url
        else:
            self.server = url + "/"

        self._auths = [("anonymous", None)]
        self._auth_resolved = False
        if username is not None and password is not None:
            self._auths[0] = (
                "basic",
                requests.auth.HTTPBasicAuth(username.encode("utf-8"), password.encode("utf-8")),
            )

        if requests_kerberos is not None:
            self._auths.append(("kerberos", requests_kerberos.HTTPKerberosAuth()))

        self.auth = None
        self.crumb = None
        self.timeout = timeout
        self._session = WrappedSession()

        extra_headers = os.environ.get("JENKINS_API_EXTRA_HEADERS", "")
        if extra_headers:
            logging.warning(
                "JENKINS_API_EXTRA_HEADERS adds these HTTP headers: %s",
                extra_headers.split("\n"),
            )
        for token in extra_headers.split("\n"):
            if ":" in token:
                header, value = token.split(":", 1)
                self._session.headers[header] = value.strip()

        if os.getenv("PYTHONHTTPSVERIFY", "1") == "0":
            logging.debug(
                "PYTHONHTTPSVERIFY=0 detected so we will "
                "disable requests library SSL verification to keep "
                "compatibility with older versions.",
            )
            requests.packages.urllib3.disable_warnings()
            self._session.verify = False

    def getEncodedParams(self, params):
        for k, v in params.items():
            if k in [
                "name",
                "msg",
                "short_name",
                "from_short_name",
                "to_short_name",
                "folder_url",
                "from_folder_url",
                "to_folder_url",
            ]:
                params[k] = quote(v.encode("utf8"))
        return params

    def buildUrl(self, format_spec, variables=None):
        url_path = format_spec % self.getEncodedParams(variables) if variables else format_spec

        return str(parse.urljoin(self.server, url_path))

    def maybeAddCrumb(self, req):
        # We don't know yet whether we need a crumb
        if self.crumb is None:
            try:
                response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(CRUMB_URL)), add_crumb=False)
                if not response:
                    raise EmptyResponseException("Empty response for crumb")
            except (NotFoundException, EmptyResponseException):
                self.crumb = False
            else:
                self.crumb = json.loads(response)
        if self.crumb:
            req.headers[self.crumb["crumbRequestField"]] = self.crumb["crumb"]

    def maybeAddAuth(self):
        if self._auth_resolved:
            return

        if len(self._auths) == 1:
            # If we only have one auth mechanism specified, just require it
            self._session.auth = self._auths[0][1]
        else:
            # Attempt the list of auth mechanisms and keep the first that works
            # otherwise default to the first one in the list (last popped).
            # This is a hack to allow the transparent use of kerberos to work
            # in future, we should require explicit request to use kerberos
            failures = []
            for name, auth in reversed(self._auths):
                try:
                    self.jenkinsOpen(
                        requests.Request("GET", self.buildUrl(INFO), auth=auth),
                        add_crumb=False,
                        resolve_auth=False,
                    )
                    self._session.auth = auth
                    break
                except TimeoutException:
                    raise
                except Exception as exc:
                    # assume authentication failure
                    failures.append("auth(%s) %s" % (name, exc))
                    continue
            else:
                raise JenkinsException("Unable to authenticate with any scheme:\n%s" % "\n".join(failures))

        self._auth_resolved = True
        self.auth = self._session.auth

    def addMissingBuilds(self, data):
        """Query Jenkins to get all builds of a job.

        The Jenkins API only fetches the first 100 builds, with no
        indicator that there are more to be fetched. This fetches more
        builds where necessary to get all builds of a given job.

        Much of this code borrowed from
        which is MIT licensed.
        """
        if not data.get("builds"):
            return data
        oldest_loaded_build_number = data["builds"][-1]["number"]
        first_build_number = oldest_loaded_build_number if not data["firstBuild"] else data["firstBuild"]["number"]
        all_builds_loaded = oldest_loaded_build_number == first_build_number
        if all_builds_loaded:
            return data
        folder_url, short_name = self.getJobFolder(data["fullName"])
        response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(ALL_BUILDS, locals())))
        if response:
            data["builds"] = json.loads(response)["allBuilds"]
        else:
            raise JenkinsException("Could not fetch all builds from job[%s]" % data["fullName"])
        return data

    def getJobInfo(self, name, depth=0, fetch_all_builds=False):
        """Get job information dictionary.

        :param name: Job name, ``str``
        :param depth: JSON depth, ``int``
        :param fetch_all_builds: If true, all builds will be retrieved
                                 from Jenkins. Otherwise, Jenkins will
                                 only return the most recent 100
                                 builds. This comes at the expense of
                                 an additional API call which may
                                 return significant amounts of
                                 data. ``bool``
        :returns: dictionary of job information
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(JOB_INFO, locals())))
            if response:
                if fetch_all_builds:
                    return self.addMissingBuilds(json.loads(response))
                else:
                    return json.loads(response)
            else:
                raise JenkinsException("job[%s] does not exist" % name)
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException("job[%s] does not exist" % name)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for job[%s]" % name)

    def getJobInfoRegex(self, pattern, depth=0, folder_depth=0, folder_depth_per_request=10):
        """Get a list of jobs information that contain names which match the
           regex pattern.

        :param pattern: regex pattern, ``str``
        :param depth: JSON depth, ``int``
        :param folder_depth: folder level depth to search ``int``
        :param folder_depth_per_request: Number of levels to fetch at once,
            ``int``. See :func:`getAllJobs`.
        :returns: List of jobs info, ``list``
        """
        result = []
        jobs = self.getAllJobs(folder_depth=folder_depth, folder_depth_per_request=folder_depth_per_request)
        for job in jobs:
            if re.search(pattern, job["name"]):
                result.append(self.getJobInfo(job["name"], depth=depth))

        return result

    def getJobName(self, name):
        """Return the name of a job using the API.

        That is roughly an identity method which can be used to quickly verify
        a job exists or is accessible without causing too much stress on the
        server side.

        :param name: Job name, ``str``
        :returns: Name of job or None
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(JOB_NAME, locals())))
        except NotFoundException:
            return None
        else:
            actual = json.loads(response)["name"]
            if actual != short_name:
                raise JenkinsException("Jenkins returned an unexpected job name %s " "(expected: %s)" % (actual, name))
            return actual

    def debugJobInfo(self, job_name):
        """Print out job info in more readable format."""
        for k, v in self.getJobInfo(job_name).items():
            print(k, v)

    def responseHandler(self, response):
        """Handle response objects"""

        # raise exceptions if occurred
        response.raise_for_status()

        # Response objects will automatically return unicode encoded
        # when accessing .text property
        return response

    def _request(self, req):
        r = self._session.prepare_request(req)
        # requests.Session.send() does not honor env settings by design
        _settings = self._session.mergeEnvSettings(r.url, {}, None, self._session.verify, None)
        _settings["timeout"] = self.timeout
        return self._session.send(r, **_settings)

    def jenkinsOpen(self, req, add_crumb=True, resolve_auth=True):
        """Return the HTTP response body from a ``requests.Request``.

        :returns: ``str``
        """
        return self.jenkinsRequest(req, add_crumb, resolve_auth).text

    def jenkinsRequest(self, req, add_crumb=True, resolve_auth=True):
        """Utility routine for opening an HTTP request to a Jenkins server.

        :param req: A ``requests.Request`` to submit.
        :param add_crumb: If True, try to add a crumb header to this ``req``
                          before submitting. Defaults to ``True``.
        :param resolve_auth: If True, maybe add authentication. Defaults to
                             ``True``.
        :returns: A ``requests.Response`` object.
        """
        try:
            if resolve_auth:
                self.maybeAddAuth()
            if add_crumb:
                self.maybeAddCrumb(req)

            return self.responseHandler(self._request(req))

        except req_exc.HTTPError as e:
            # Jenkins's funky authentication means its nigh impossible to
            # distinguish errors.
            if e.response.status_code in [401, 403, 500]:
                msg = "Error in request. " + "Possibly authentication failed [%s]: %s" % (
                    e.response.status_code,
                    e.response.reason,
                )
                if e.response.text:
                    msg += "\n" + e.response.text
                raise JenkinsException(msg)
            elif e.response.status_code == 404:
                raise NotFoundException("Requested item could not be found")
            else:
                raise
        except req_exc.Timeout as e:
            raise TimeoutException("Error in request: %s" % (e))
        except ConnectionError as e:
            # python 2.6 compatibility to ensure same exception raised
            # since URLError wraps a socket timeout on python 2.6.
            if str(e.reason) == "timed out":
                raise TimeoutException("Error in request: %s" % (e.reason))
            raise JenkinsException("Error in request: %s" % (e.reason))

    def getQueueItem(self, number, depth=0):
        """Get information about a queued item (to-be-created job).

        The returned dict will have a "why" key if the queued item is still
        waiting for an executor.

        The returned dict will have an "executable" key if the queued item is
        running on an executor, or has completed running. Use this to
        determine the job number / URL.

        :param name: queue number, ``int``
        :returns: dictionary of queued information, ``dict``
        """
        url = self.buildUrl(Q_ITEM, locals())
        try:
            response = self.jenkinsOpen(requests.Request("GET", url))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("queue number[%d] does not exist" % number)
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException("queue number[%d] does not exist" % number)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for queue number[%d]" % number)

    def getBuildInfo(self, name, number, depth=0):
        """Get build information dictionary.

        :param name: Job name, ``str``
        :param number: Build number, ``int``
        :param depth: JSON depth, ``int``
        :returns: dictionary of build information, ``dict``

        Example::
            >>> next_build_number = server.getJobInfo('build_name')['nextBuildNumber']
            >>> output = server.buildJob('build_name')
            >>> from time import sleep; sleep(10)
            >>> build_info = server.getBuildInfo('build_name', next_build_number)
            >>> print(build_info)
            {u'building': False, u'changeSet': {u'items': [{u'date': u'2011-12-19T18:01:52.540557Z', u'msg': u'test', u'revision': 66, u'user': u'unknown', u'paths': [{u'editType': u'edit', u'file': u'/branches/demo/index.html'}]}], u'kind': u'svn', u'revisions': [{u'module': u'http://eaas-svn01.i3.level3.com/eaas', u'revision': 66}]}, u'builtOn': u'', u'description': None, u'artifacts': [{u'relativePath': u'dist/eaas-87-2011-12-19_18-01-57.war', u'displayPath': u'eaas-87-2011-12-19_18-01-57.war', u'fileName': u'eaas-87-2011-12-19_18-01-57.war'}, {u'relativePath': u'dist/eaas-87-2011-12-19_18-01-57.war.zip', u'displayPath': u'eaas-87-2011-12-19_18-01-57.war.zip', u'fileName': u'eaas-87-2011-12-19_18-01-57.war.zip'}], u'timestamp': 1324317717000, u'number': 87, u'actions': [{u'parameters': [{u'name': u'SERVICE_NAME', u'value': u'eaas'}, {u'name': u'PROJECT_NAME', u'value': u'demo'}]}, {u'causes': [{u'userName': u'anonymous', u'shortDescription': u'Started by user anonymous'}]}, {}, {}, {}], u'id': u'2011-12-19_18-01-57', u'keepLog': False, u'url': u'http://eaas-jenkins01.i3.level3.com:9080/job/build_war/87/', u'culprits': [{u'absoluteUrl': u'http://eaas-jenkins01.i3.level3.com:9080/user/unknown', u'fullName': u'unknown'}], u'result': u'SUCCESS', u'duration': 8826, u'fullDisplayName': u'build_war #87'}
        """  # noqa: E501
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(BUILD_INFO, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except ValueError:
            raise JenkinsException("Could not parse JSON info for job[%s] number[%d]" % (name, number))

    def getBuildEnvVars(self, name, number, depth=0):
        """Get build environment variables.

        :param name: Job name, ``str``
        :param number: Build number, ``int``
        :param depth: JSON depth, ``int``
        :returns: dictionary of build env vars, ``dict`` or None for workflow jobs,
            or if InjectEnvVars plugin not installed
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(BUILD_ENV_VARS, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except req_exc.HTTPError:
            raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except ValueError:
            raise JenkinsException("Could not parse JSON info for job[%s] number[%d]" % (name, number))
        except NotFoundException:
            # This can happen on workflow jobs, or if InjectEnvVars plugin not installed
            return None

    def getBuildTestReport(self, name, number, depth=0):
        """Get test results report.

        :param name: Job name, ``str``
        :param number: Build number, ``int``
        :returns: dictionary of test report results, ``dict`` or None if there is no Test Report
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(BUILD_TEST_REPORT, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except req_exc.HTTPError:
            raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except ValueError:
            raise JenkinsException("Could not parse JSON info for job[%s] number[%d]" % (name, number))
        except NotFoundException:
            # This can happen if the test report wasn't generated for any reason
            return None

    def getQueueInfo(self):
        """:returns: list of job dictionaries, ``[dict]``

        Example::
            >>> queue_info = server.getQueueInfo()
            >>> print(queue_info[0])
            {u'task': {u'url': u'http://your_url/job/my_job/', u'color': u'aborted_anime', u'name': u'my_job'}, u'stuck': False, u'actions': [{u'causes': [{u'shortDescription': u'Started by timer'}]}], u'buildable': False, u'params': u'', u'buildableStartMilliseconds': 1315087293316, u'why': u'Build #2,532 is already in progress (ETA:10 min)', u'blocked': True}
        """  # noqa: E501
        return json.loads(self.jenkinsOpen(requests.Request("GET", self.buildUrl(Q_INFO))))["items"]

    def cancelQueue(self, id):
        """Cancel a queued build.

        :param id: Jenkins job id number for the build, ``int``
        """
        # Jenkins seems to always return a 404 when using this REST endpoint
        # https://issues.jenkins-ci.org/browse/JENKINS-21311
        with contextlib.suppress(NotFoundException):
            self.jenkinsOpen(
                requests.Request(
                    "POST",
                    self.buildUrl(CANCEL_QUEUE, locals()),
                    headers={"Referer": self.server},
                ),
            )

    def getInfo(self, item="", query=None):
        """Get information on this Master or item on Master.

        This information includes job list and view information and can be
        used to retreive information on items such as job folders.

        :param item: item to get information about on this Master
        :param query: xpath to extract information about on this Master
        :returns: dictionary of information about Master or item, ``dict``

        Example::

            >>> info = server.getInfo()
            >>> jobs = info['jobs']
            >>> print(jobs[0])
            {u'url': u'http://your_url_here/job/my_job/', u'color': u'blue',
            u'name': u'my_job'}

        """
        url = "/".join((item, INFO)).lstrip("/")
        url = quote(url)
        if query:
            url += query
        try:
            return json.loads(self.jenkinsOpen(requests.Request("GET", self.buildUrl(url))))
        except req_exc.HTTPError:
            raise BadHTTPException("Error communicating with server[%s]" % self.server)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for server[%s]" % self.server)

    def getWhoami(self, depth=0):
        """Get information about the user account that authenticated to
        Jenkins. This is a simple way to verify that your credentials are
        correct.

        :returns: Information about the current user ``dict``

        Example::

            >>> me = server.getWhoami()
            >>> print me['fullName']
            >>> 'John'

        """
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(WHOAMI_URL, locals())))
            if response is None:
                raise EmptyResponseException("Error communicating with server[%s]: " "empty response" % self.server)

            return json.loads(response)

        except req_exc.HTTPError:
            raise BadHTTPException("Error communicating with server[%s]" % self.server)

    def getVersion(self):
        """Get the version of this Master.

        :returns: This master's version number ``str``

        Example::

            >>> info = server.getVersion()
            >>> print (info)
            >>> 1.541

        """
        try:
            request = requests.Request("GET", self.buildUrl(""))
            request.headers["X-Jenkins"] = "0.0"
            response = self.responseHandler(self._request(request))

            return response.headers["X-Jenkins"]

        except req_exc.HTTPError:
            raise BadHTTPException("Error communicating with server[%s]" % self.server)

    def getPluginsInfo(self, depth=2):
        """Get all installed plugins information on this Master.

        This method retrieves information about each plugin that is installed
        on master returning the raw plugin data in a JSON format.

        .. deprecated:: 0.4.9
           Use :func:`getPlugin` instead.

        :param depth: JSON depth, ``int``
        :returns: info on all plugins ``[dict]``

        Example::

            >>> info = server.getPluginsInfo()
            >>> print(info)
            [{u'backupVersion': None, u'version': u'0.0.4', u'deleted': False,
            u'supportsDynamicLoad': u'MAYBE', u'hasUpdate': True,
            u'enabled': True, u'pinned': False, u'downgradable': False,
            u'dependencies': [], u'url':
            u'http://wiki.jenkins-ci.org/display/JENKINS/Gearman+Plugin',
            u'longName': u'Gearman Plugin', u'active': True, u'shortName':
            u'gearman-plugin', u'bundled': False}, ..]

        """
        warnings.warn("getPluginsInfo() is deprecated, use getPlugin()", DeprecationWarning)
        return list(self.getPlugin(depth).values())

    def getPluginInfo(self, name, depth=2):
        """Get an installed plugin information on this Master.

        This method retrieves information about a specific plugin and returns
        the raw plugin data in a JSON format.
        The passed in plugin name (short or long) must be an exact match.

        .. note:: Calling this method will query Jenkins fresh for the
            information for all plugins on each call. If you need to retrieve
            information for multiple plugins it's recommended to use
            :func:`getPlugin` instead, which will return a multi key
            dictionary that can be accessed via either the short or long name
            of the plugin.

        :param name: Name (short or long) of plugin, ``str``
        :param depth: JSON depth, ``int``
        :returns: a specific plugin ``dict``

        Example::

            >>> info = server.getPluginInfo("Gearman Plugin")
            >>> print(info)
            {u'backupVersion': None, u'version': u'0.0.4', u'deleted': False,
            u'supportsDynamicLoad': u'MAYBE', u'hasUpdate': True,
            u'enabled': True, u'pinned': False, u'downgradable': False,
            u'dependencies': [], u'url':
            u'http://wiki.jenkins-ci.org/display/JENKINS/Gearman+Plugin',
            u'longName': u'Gearman Plugin', u'active': True, u'shortName':
            u'gearman-plugin', u'bundled': False}

        """
        plugins_info = self.getPlugin(depth)
        try:
            return plugins_info[name]
        except KeyError:
            pass

    def getPlugin(self, depth=2):
        """Return plugins info using helper class for version comparison

        This method retrieves information about all the installed plugins and
        uses a Plugin helper class to simplify version comparison. Also uses
        a multi key dict to allow retrieval via either short or long names.

        When printing/dumping the data, the version will transparently return
        a unicode string, which is exactly what was previously returned by the
        API.

        :param depth: JSON depth, ``int``
        :returns: info on all plugins ``[dict]``

        Example::

            >>> j = Jenkins()
            >>> info = j.getPlugin()
            >>> print(info)
            {('gearman-plugin', 'Gearman Plugin'):
              {u'backupVersion': None, u'version': u'0.0.4',
               u'deleted': False, u'supportsDynamicLoad': u'MAYBE',
               u'hasUpdate': True, u'enabled': True, u'pinned': False,
               u'downgradable': False, u'dependencies': [], u'url':
               u'http://wiki.jenkins-ci.org/display/JENKINS/Gearman+Plugin',
               u'longName': u'Gearman Plugin', u'active': True, u'shortName':
               u'gearman-plugin', u'bundled': False}, ...}

        """

        try:
            plugins_info_json = json.loads(
                self.jenkinsOpen(requests.Request("GET", self.buildUrl(PLUGIN_INFO, locals()))),
            )
        except req_exc.HTTPError:
            raise BadHTTPException("Error communicating with server[%s]" % self.server)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for server[%s]" % self.server)

        plugins_data = MultiKeyDict()
        for plugin_data in plugins_info_json["plugins"]:
            keys = (str(plugin_data["shortName"]), str(plugin_data["longName"]))
            plugins_data[keys] = Plugin(**plugin_data)

        return plugins_data

    def getJobs(self, folder_depth=0, folder_depth_per_request=10, view_name=None):
        """Get list of jobs.

        Each job is a dictionary with 'name', 'url', 'color' and 'fullname'
        keys.

        If the ``view_name`` parameter is present, the list of
        jobs will be limited to only those configured in the
        specified view. In this case, the job dictionary 'fullname' key
        would be equal to the job name.

        :param folder_depth: Number of levels to search, ``int``. By default
            0, which will limit search to toplevel. None disables the limit.
        :param folder_depth_per_request: Number of levels to fetch at once,
            ``int``. See :func:`getAllJobs`.
        :param view_name: Name of a Jenkins view for which to
            retrieve jobs, ``str``. By default, the job list is
            not limited to a specific view.
        :returns: list of jobs, ``[{str: str, str: str, str: str, str: str}]``

        Example::

            >>> jobs = server.getJobs()
            >>> print(jobs)
            [{
                u'name': u'all_tests',
                u'url': u'http://your_url.here/job/all_tests/',
                u'color': u'blue',
                u'fullname': u'all_tests'
            }]

        """

        if view_name:
            return self.getViewJobs(name=view_name)
        else:
            return self.getAllJobs(
                folder_depth=folder_depth,
                folder_depth_per_request=folder_depth_per_request,
            )

    def getAllJobs(self, folder_depth=None, folder_depth_per_request=10):
        """Get list of all jobs recursively to the given folder depth.

        Each job is a dictionary with 'name', 'url', 'color' and 'fullname'
        keys.

        :param folder_depth: Number of levels to search, ``int``. By default
            None, which will search all levels. 0 limits to toplevel.
        :param folder_depth_per_request: Number of levels to fetch at once,
            ``int``. By default 10, which is usually enough to fetch all jobs
            using a single request and still easily fits into an HTTP request.
        :returns: list of jobs, ``[ { str: str} ]``

        .. note::

            On instances with many folders it would not be efficient to fetch
            each folder separately, hence `folder_depth_per_request` levels
            are fetched at once using the ``tree`` query parameter::

                ?tree=jobs[url,color,name,jobs[...,jobs[...,jobs[...,jobs]]]]

            If there are more folder levels than the query asks for, Jenkins
            returns empty [#]_ objects at the deepest level::

                {"name": "folder", "url": "...", "jobs": [{}, {}, ...]}

            This makes it possible to detect when additional requests are
            needed.

            .. [#] Actually recent Jenkins includes a ``_class`` field
                everywhere, but it's missing the requested fields.
        """
        jobs_query = "jobs"
        for _ in range(folder_depth_per_request):
            jobs_query = JOBS_QUERY_TREE % jobs_query
        jobs_query = JOBS_QUERY % jobs_query

        jobs_list = []
        jobs = [(0, [], self.getInfo(query=jobs_query)["jobs"])]
        for lvl, root, lvl_jobs in jobs:
            if not isinstance(lvl_jobs, list):
                lvl_jobs = [lvl_jobs]
            for job in lvl_jobs:
                path = [*root, job["name"]]
                # insert fullname info if it doesn't exist to
                # allow callers to easily reference unambiguously
                if "fullname" not in job:
                    job["fullname"] = "/".join(path)
                jobs_list.append(job)
                if "jobs" in job and isinstance(job["jobs"], list):  # folder
                    if folder_depth is None or lvl < folder_depth:
                        children = job["jobs"]
                        # once folder_depth_per_request is reached, Jenkins
                        # returns empty objects
                        if any("url" not in child for child in job["jobs"]):
                            url_path = "".join(["/job/" + p for p in path])
                            children = self.getInfo(url_path, query=jobs_query)["jobs"]
                        jobs.append((lvl + 1, path, children))
        return jobs_list

    def copyJob(self, from_name, to_name):
        """Copy a Jenkins job.

        Will raise an exception whenever the source and destination folder
        for this jobs won't be the same.

        :param from_name: Name of Jenkins job to copy from, ``str``
        :param to_name: Name of Jenkins job to copy to, ``str``
        :throws: :class:`JenkinsException` whenever the source and destination
            folder are not the same
        """
        from_folder_url, from_short_name = self.getJobFolder(from_name)
        to_folder_url, to_short_name = self.getJobFolder(to_name)
        if from_folder_url != to_folder_url:
            raise JenkinsException(
                "copy[%s to %s] failed, source and destination " "folder must be the same" % (from_name, to_name),
            )

        self.jenkinsOpen(requests.Request("POST", self.buildUrl(COPY_JOB, locals())))
        self.assertJobExists(to_name, "create[%s] failed")

    def renameJob(self, from_name, to_name):
        """Rename an existing Jenkins job

        Will raise an exception whenever the source and destination folder
        for this jobs won't be the same.

        :param from_name: Name of Jenkins job to rename, ``str``
        :param to_name: New Jenkins job name, ``str``
        :throws: :class:`JenkinsException` whenever the source and destination
            folder are not the same
        """
        from_folder_url, from_short_name = self.getJobFolder(from_name)
        to_folder_url, to_short_name = self.getJobFolder(to_name)
        if from_folder_url != to_folder_url:
            raise JenkinsException(
                "rename[%s to %s] failed, source and destination folder " "must be the same" % (from_name, to_name),
            )
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(RENAME_JOB, locals())))
        self.assertJobExists(to_name, "rename[%s] failed")

    def deleteJob(self, name):
        """Delete Jenkins job permanently.

        :param name: Name of Jenkins job, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DELETE_JOB, locals())))
        if self.jobExists(name):
            raise JenkinsException("delete[%s] failed" % (name))

    def enableJob(self, name):
        """Enable Jenkins job.

        :param name: Name of Jenkins job, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(ENABLE_JOB, locals())))

    def disableJob(self, name):
        """Disable Jenkins job.

        To re-enable, call :meth:`Jenkins.enableJob`.

        :param name: Name of Jenkins job, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DISABLE_JOB, locals())))

    def setNextbBuildNumber(self, name, number):
        """Set a job's next build number.

        The current next build number is contained within the job
        information retrieved using :meth:`Jenkins.getJobInfo`.  If
        the specified next build number is less than the last build
        number, Jenkins will ignore the request.

        Note that the `Next Build Number Plugin
        <https://wiki.jenkins-ci.org/display/JENKINS/Next+Build+Number+Plugin>`_
        must be installed to enable this functionality.

        :param name: Name of Jenkins job, ``str``
        :param number: Next build number to set, ``int``

        Example::

            >>> next_bn = server.getJobInfo('job_name')['nextBuildNumber']
            >>> server.setNextbBuildNumber('job_name', next_bn + 50)
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(
            requests.Request(
                "POST",
                self.buildUrl(SET_JOB_BUILD_NUMBER, locals()),
                data=("nextBuildNumber=%d" % number).encode("utf-8"),
            ),
        )

    def jobExists(self, name):
        """Check whether a job exists

        :param name: Name of Jenkins job, ``str``
        :returns: ``True`` if Jenkins job exists
        """
        folder_url, short_name = self.getJobFolder(name)
        if self.getJobName(name) == short_name:
            return True
        return None

    def jobs_count(self):
        """Get the number of jobs on the Jenkins server

        :returns: Total number of jobs, ``int``
        """
        return len(self.getAllJobs())

    def assertJobExists(self, name, exception_message="job[%s] does not exist"):
        """Raise an exception if a job does not exist

        :param name: Name of Jenkins job, ``str``
        :param exception_message: Message to use for the exception. Formatted
                                  with ``name``
        :throws: :class:`JenkinsException` whenever the job does not exist
        """
        if not self.jobExists(name):
            raise JenkinsException(exception_message % name)

    def createFolder(self, folder_name, ignore_failures=False):
        """Create a new Jenkins folder

        :param folder_name: Name of Jenkins Folder, ``str``
        :param ignore_failures: if True, don't raise if it was not possible to create the folder, ``bool``
        """
        folder_url, short_name = self.getJobFolder(folder_name)
        url = self.buildUrl(CREATE_JOB, locals())
        data = {
            "name": folder_name,
            "mode": "com.cloudbees.hudson.plugins.folder.Folder",
        }
        try:
            response = self.jenkinsRequest(requests.Request("POST", url, data=data))
        except requests.exceptions.HTTPError:
            if not ignore_failures:
                raise JenkinsException("Error creating folder [%s]. Probably it already exists." % (folder_name))

    def upsertJob(self, name, config_xml):
        """Create a new Jenkins job or reconfigures it if it exists

        :param name: Name of Jenkins job, ``str``
        :param config_xml: config file text, ``str``
        """

        if self.jobExists(name):
            self.reconfigJob(name, config_xml)
        else:
            self.createJob(name, config_xml)

    def checkJenkinsfileSyntax(self, jenkinsfile):
        """Checks if a Pipeline Jenkinsfile has a valid syntax

        :param jenkinsfile: Jenkinsfile text, ``str``
        :returns: List of errors in the Jenkinsfile. Empty list if no errors.
        """
        # https://jenkins.io/doc/book/pipeline/development/#linter
        # JENKINS_CRUMB=`curl "$JENKINS_URL/crumbIssuer/api/xml?xpath=concat(//crumbRequestField,\":\",//crumb)"
        # curl -X POST -H $JENKINS_CRUMB -F "jenkinsfile=<Jenkinsfile" $JENKINS_URL/pipeline-model-converter/val
        url = self.buildUrl(CHECK_JENKINSFILE_SYNTAX, locals())
        the_data = {"jenkinsfile": jenkinsfile}
        response = self.jenkinsRequest(requests.Request("POST", url, data=the_data))
        return response.json().get("data", {}).get("errors", [])

    def createJob(self, name, config_xml):
        """Create a new Jenkins job

        :param name: Name of Jenkins job, ``str``
        :param config_xml: config file text, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        if self.jobExists(name):
            raise JenkinsException("job[%s] already exists" % (name))

        try:
            self.jenkinsOpen(
                requests.Request(
                    "POST",
                    self.buildUrl(CREATE_JOB, locals()),
                    data=config_xml.encode("utf-8"),
                    headers=DEFAULT_HEADERS,
                ),
            )
        except NotFoundException:
            raise JenkinsException("Cannot create job[%s] because folder " "for the job does not exist" % (name))
        self.assertJobExists(name, "create[%s] failed")

    def getJobConfig(self, name):
        """Get configuration of existing Jenkins job.

        :param name: Name of Jenkins job, ``str``
        :returns: job configuration (XML format)
        """
        folder_url, short_name = self.getJobFolder(name)
        request = requests.Request("GET", self.buildUrl(CONFIG_JOB, locals()))
        return self.jenkinsOpen(request)

    def reconfigJob(self, name, config_xml):
        """Change configuration of existing Jenkins job.

        To create a new job, see :meth:`Jenkins.createJob`.

        :param name: Name of Jenkins job, ``str``
        :param config_xml: New XML configuration, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        reconfig_url = self.buildUrl(CONFIG_JOB, locals())
        self.jenkinsOpen(
            requests.Request(
                "POST",
                reconfig_url,
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )

    def buildJobUrl(self, name, parameters=None, token=None):
        """Get URL to trigger build job.

        Authenticated setups may require configuring a token on the server
        side.

        Use ``list of two membered tuples`` to supply parameters with multi
        select options.

        :param name: Name of Jenkins job, ``str``
        :param parameters: parameters for job, or None., ``dict`` or
            ``list of two membered tuples``
        :param token: (optional) token for building job, ``str``
        :returns: URL for building job
        """
        folder_url, short_name = self.getJobFolder(name)
        if parameters:
            if token:
                if isinstance(parameters, list):
                    parameters.append(
                        (
                            "token",
                            token,
                        ),
                    )
                elif isinstance(parameters, dict):
                    parameters.update({"token": token})
                else:
                    raise JenkinsException(
                        "build parameters can be a dictionary "
                        'like {"param_key": "param_value", ...} '
                        "or a list of two membered tuples "
                        'like [("param_key", "param_value",), ...]',
                    )
            return self.buildUrl(BUILD_WITH_PARAMS_JOB, locals()) + "?" + parse.urlencode(parameters)
        elif token:
            return self.buildUrl(BUILD_JOB, locals()) + "?" + parse.urlencode({"token": token})
        else:
            return self.buildUrl(BUILD_JOB, locals())

    def buildJob(self, name, parameters=None, token=None):
        """Trigger build job.

        This method returns a queue item number that you can pass to
        :meth:`Jenkins.getQueueItem`. Note that this queue number is only
        valid for about five minutes after the job completes, so you should
        get/poll the queue information as soon as possible to determine the
        job's URL.

        :param name: name of job
        :param parameters: parameters for job, or ``None``, ``dict``
        :param token: Jenkins API token
        :returns: ``int`` queue item
        """
        response = self.jenkinsRequest(requests.Request("POST", self.buildJobUrl(name, parameters, token)))

        if "Location" not in response.headers:
            raise EmptyResponseException("Header 'Location' not found in " "response from server[%s]" % self.server)

        location = response.headers["Location"]
        # location is a queue item, eg. "http://jenkins/queue/item/25/"
        if location.endswith("/"):
            location = location[:-1]
        parts = location.split("/")
        number = int(parts[-1])
        return number

    def runScript(self, script, node=None):
        """Execute a groovy script on the jenkins master or on a node if
        specified..

        :param script: The groovy script, ``string``
        :param node: Node to run the script on, defaults to None (master).
        :returns: The result of the script run.

        Example::
            >>> info = server.runScript("println(Jenkins.instance.pluginManager.plugins)")
            >>> print(info)
            u'[Plugin:windows-slaves, Plugin:ssh-slaves, Plugin:translation,
            Plugin:cvs, Plugin:nodelabelparameter, Plugin:external-monitor-job,
            Plugin:mailer, Plugin:jquery, Plugin:antisamy-markup-formatter,
            Plugin:maven-plugin, Plugin:pam-auth]'
        """
        magic_str = ")]}."
        print_magic_str = 'print("{}")'.format(magic_str)
        data = {"script": "{0}\n{1}".format(script, print_magic_str).encode("utf-8")}
        url = self.buildUrl(NODE_SCRIPT_TEXT, locals()) if node else self.buildUrl(SCRIPT_TEXT, locals())

        result = self.jenkinsOpen(requests.Request("POST", url, data=data))

        if not result.endswith(magic_str):
            raise JenkinsException(result)

        return result[: result.rfind("\n")]

    def installPlugin(self, name, include_dependencies=True):
        """Install a plugin and its dependencies from the Jenkins public
        repository at http://repo.jenkins-ci.org/repo/org/jenkins-ci/plugins

        :param name: The plugin short name, ``string``
        :param include_dependencies: Install the plugin's dependencies, ``bool``
        :returns: Whether a Jenkins restart is required, ``bool``

        Example::
            >>> info = server.installPlugin("jabber")
            >>> print(info)
            True
        """
        # using a groovy script because Jenkins does not provide a REST endpoint
        # for installing plugins.
        install = 'Jenkins.instance.updateCenter.getPlugin("' + name + '")' ".deploy();"
        if include_dependencies:
            install = (
                'Jenkins.instance.updateCenter.getPlugin("' + name + '")' ".getNeededDependencies().each{it.deploy()};"
            ) + install

        self.runScript(install)
        # runScript is an async call to run groovy. we need to wait a little
        # before we can get a reliable response on whether a restart is needed
        time.sleep(2)
        is_restart_required = "Jenkins.instance.updateCenter" ".isRestartRequiredForCompletion()"

        # response is a string (i.e. u'Result: true\n'), return a bool instead
        response_str = self.runScript(is_restart_required)
        response = response_str.split(":")[1].strip().lower() == "true"
        return response

    def stopBuild(self, name, number):
        """Stop a running Jenkins build.

        :param name: Name of Jenkins job, ``str``
        :param number: Jenkins build number for the job, ``int``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(STOP_BUILD, locals())))

    def deleteBuild(self, name, number):
        """Delete a Jenkins build.

        :param name: Name of Jenkins job, ``str``
        :param number: Jenkins build number for the job, ``int``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DELETE_BUILD, locals()), b""))

    def wipeoutJobWorkspace(self, name):
        """Wipe out workspace for given Jenkins job.

        :param name: Name of Jenkins job, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(WIPEOUT_JOB_WORKSPACE, locals()), b""))

    def getRunningBuilds(self):
        """Return list of running builds.

        Each build is a dict with keys 'name', 'number', 'url', 'node',
        and 'executor'.

        :returns: List of builds,
          ``[ { str: str, str: int, str:str, str: str, str: int} ]``

        Example::
            >>> builds = server.getRunningBuilds()
            >>> print(builds)
            [{'node': 'foo-slave', 'url': 'https://localhost/job/test/15/',
              'executor': 0, 'name': 'test', 'number': 15}]
        """
        builds = []
        nodes = self.getNodes()
        for node in nodes:
            # the name returned is not the name to lookup when
            # dealing with master :/
            node_name = "(master)" if node["name"] == "master" else node["name"]
            try:
                info = self.getNodeInfo(node_name, depth=2)
            except JenkinsException as e:
                # Jenkins may 500 on depth >0. If the node info comes back
                # at depth 0 treat it as a node not running any jobs.
                if "[500]" in str(e) and self.getNodeInfo(node_name, depth=0):
                    continue
                else:
                    raise
            for executor in info["executors"]:
                executable = executor["currentExecutable"]
                if executable and "number" in executable:
                    executor_number = executor["number"]
                    build_number = executable["number"]
                    url = executable["url"]
                    m = re.search(r"/job/([^/]+)/.*", parse.urlparse(url).path)
                    job_name = m.group(1)
                    builds.append(
                        {
                            "name": job_name,
                            "number": build_number,
                            "url": url,
                            "node": node_name,
                            "executor": executor_number,
                        },
                    )
        return builds

    def getNodes(self, depth=0):
        """Get a list of nodes connected to the Master

        Each node is a dict with keys 'name' and 'offline'

        :returns: List of nodes, ``[ { str: str, str: bool} ]``
        """
        try:
            nodes_data = json.loads(self.jenkinsOpen(requests.Request("GET", self.buildUrl(NODE_LIST, locals()))))
            return [{"name": c["displayName"], "offline": c["offline"]} for c in nodes_data["computer"]]
        except req_exc.HTTPError:
            raise BadHTTPException("Error communicating with server[%s]" % self.server)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for server[%s]" % self.server)

    def getNodeInfo(self, name, depth=0):
        """Get node information dictionary

        :param name: Node name, ``str``
        :param depth: JSON depth, ``int``
        :returns: Dictionary of node info, ``dict``
        """
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(NODE_INFO, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("node[%s] does not exist" % name)
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException("node[%s] does not exist" % name)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for node[%s]" % name)

    def nodeExists(self, name):
        """Check whether a node exists

        :param name: Name of Jenkins node, ``str``
        :returns: ``True`` if Jenkins node exists
        """
        try:
            self.getNodeInfo(name)
            return True
        except JenkinsException:
            return False

    def assertNodeExists(self, name, exception_message="node[%s] does not exist"):
        """Raise an exception if a node does not exist

        :param name: Name of Jenkins node, ``str``
        :param exception_message: Message to use for the exception. Formatted
                                  with ``name``
        :throws: :class:`JenkinsException` whenever the node does not exist
        """
        if not self.nodeExists(name):
            raise JenkinsException(exception_message % name)

    def deleteNode(self, name):
        """Delete Jenkins node permanently.

        :param name: Name of Jenkins node, ``str``
        """
        self.getNodeInfo(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DELETE_NODE, locals())))
        if self.nodeExists(name):
            raise JenkinsException("delete[%s] failed" % (name))

    def disableNode(self, name, msg=""):
        """Disable a node

        :param name: Jenkins node name, ``str``
        :param msg: Offline message, ``str``
        """
        info = self.getNodeInfo(name)
        if info["offline"]:
            return
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(TOGGLE_OFFLINE, locals())))

    def enableNode(self, name):
        """Enable a node

        :param name: Jenkins node name, ``str``
        """
        info = self.getNodeInfo(name)
        if not info["offline"]:
            return
        msg = ""
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(TOGGLE_OFFLINE, locals())))

    def createNode(
        self,
        name,
        numExecutors=2,
        nodeDescription=None,
        remoteFS="/var/lib/jenkins",
        labels=None,
        exclusive=False,
        launcher=LAUNCHER_COMMAND,
        launcher_params={},
    ):
        """Create a node

        :param name: name of node to create, ``str``
        :param numExecutors: number of executors for node, ``int``
        :param nodeDescription: Description of node, ``str``
        :param remoteFS: Remote filesystem location to use, ``str``
        :param labels: Labels to associate with node, ``str``
        :param exclusive: Use this node for tied jobs only, ``bool``
        :param launcher: The launch method for the slave, ``jenkins.LAUNCHER_COMMAND``, \
        ``jenkins.LAUNCHER_SSH``, ``jenkins.LAUNCHER_JNLP``, ``jenkins.LAUNCHER_WINDOWS_SERVICE``
        :param launcher_params: Additional parameters for the launcher, ``dict``
        """
        if self.nodeExists(name):
            raise JenkinsException("node[%s] already exists" % (name))

        mode = "NORMAL"
        if exclusive:
            mode = "EXCLUSIVE"

        launcher_params["stapler-class"] = launcher

        inner_params = {
            "nodeDescription": nodeDescription,
            "numExecutors": numExecutors,
            "remoteFS": remoteFS,
            "labelString": labels,
            "mode": mode,
            "retentionStrategy": {"stapler-class": "hudson.slaves.RetentionStrategy$Always"},
            "nodeProperties": {"stapler-class-bag": "true"},
            "launcher": launcher_params,
        }

        params = {"name": name, "type": NODE_TYPE, "json": json.dumps(inner_params)}

        self.jenkinsOpen(requests.Request("POST", self.buildUrl(CREATE_NODE, locals()), data=params))

        self.assertNodeExists(name, "create[%s] failed")

    def getNodeConfig(self, name):
        """Get the configuration for a node.

        :param name: Jenkins node name, ``str``
        """
        get_config_url = self.buildUrl(CONFIG_NODE, locals())
        return self.jenkinsOpen(requests.Request("GET", get_config_url))

    def reconfigNode(self, name, config_xml):
        """Change the configuration for an existing node.

        :param name: Jenkins node name, ``str``
        :param config_xml: New XML configuration, ``str``
        """
        reconfig_url = self.buildUrl(CONFIG_NODE, locals())
        self.jenkinsOpen(
            requests.Request(
                "POST",
                reconfig_url,
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )

    def getBuildConsoleOutput(self, name, number):
        """Get build console text.

        :param name: Job name, ``str``
        :param number: Build number, ``int``
        :returns: Build console output,  ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(BUILD_CONSOLE_OUTPUT, locals())))
            if response:
                return response
            else:
                raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException("job[%s] number[%d] does not exist" % (name, number))

    def getJobFolder(self, name):
        """Return the name and folder (see cloudbees plugin).

        This is a method to support cloudbees folder plugin.
        Url request should take into account folder path when the job name specify it
        (ex.: 'folder/job')

        :param name: Job name, ``str``
        :returns: Tuple [ 'folder path for Request', 'Name of job without folder path' ]
        """

        a_path = name.split("/")
        short_name = a_path[-1]
        folder_url = ("job/" + "/job/".join(a_path[:-1]) + "/") if len(a_path) > 1 else ""

        return folder_url, short_name

    def getViewJobs(self, name):
        """Get list of jobs on the view specified.

        Each job is a dictionary with 'name', 'url', 'color' and 'fullname'
        keys.

        The list of jobs is limited to only those configured in the
        specified view. Each job dictionary 'fullname' key
        is equal to the job name.

        :param view_name: Name of a Jenkins view for which to
            retrieve jobs, ``str``.
        :returns: list of jobs, ``[{str: str, str: str, str: str, str: str}]``
        """

        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(VIEW_JOBS, locals())))
            if response:
                jobs = json.loads(response)["jobs"]
            else:
                raise JenkinsException("view[%s] does not exist" % name)
        except NotFoundException:
            raise JenkinsException("view[%s] does not exist" % name)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for view[%s]" % name)

        for job_dict in jobs:
            job_dict.update({"fullname": job_dict["name"]})

        return jobs

    def getViewName(self, name):
        """Return the name of a view using the API.

        That is roughly an identity method which can be used to quickly verify
        a view exists or is accessible without causing too much stress on the
        server side.

        :param name: View name, ``str``
        :returns: Name of view or None
        """
        folder_url, short_name = self.getJobFolder(name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(VIEW_NAME, locals())))
        except NotFoundException:
            return None
        else:
            actual = json.loads(response)["name"]
            if actual != short_name:
                raise JenkinsException(
                    "Jenkins returned an unexpected view name %s " "(expected: %s)" % (actual, short_name),
                )
            return name

    def assertViewExists(self, name, exception_message="view[%s] does not exist"):
        """Raise an exception if a view does not exist

        :param name: Name of Jenkins view, ``str``
        :param exception_message: Message to use for the exception. Formatted
                                  with ``name``
        :throws: :class:`JenkinsException` whenever the view does not exist
        """
        if not self.viewExists(name):
            raise NotFoundException(exception_message % name)

    def viewExists(self, name):
        """Check whether a view exists

        :param name: Name of Jenkins view, ``str``
        :returns: ``True`` if Jenkins view exists
        """
        if self.getViewName(name) == name:
            return True
        return None

    def viewExists(self):
        """Get list of views running.

        Each view is a dictionary with 'name' and 'url' keys.

        :returns: list of views, ``[ { str: str} ]``
        """
        return self.getInfo()["views"]

    def deleteView(self, name):
        """Delete Jenkins view permanently.

        :param name: Name of Jenkins view, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DELETE_VIEW, locals())))
        if self.viewExists(name):
            raise JenkinsException("delete[%s] failed" % (name))

    def createView(self, name, config_xml):
        """Create a new Jenkins view

        :param name: Name of Jenkins view, ``str``
        :param config_xml: config file text, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        if self.viewExists(name):
            raise JenkinsException("view[%s] already exists" % (name))

        self.jenkinsOpen(
            requests.Request(
                "POST",
                self.buildUrl(CREATE_VIEW, locals()),
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )
        self.assertViewExists(name, "create[%s] failed")

    def reconfigView(self, name, config_xml):
        """Change configuration of existing Jenkins view.

        To create a new view, see :meth:`Jenkins.createView`.

        :param name: Name of Jenkins view, ``str``
        :param config_xml: New XML configuration, ``str``
        """
        folder_url, short_name = self.getJobFolder(name)
        reconfig_url = self.buildUrl(CONFIG_VIEW, locals())
        self.jenkinsOpen(
            requests.Request(
                "POST",
                reconfig_url,
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )

    def getViewConfig(self, name):
        """Get configuration of existing Jenkins view.

        :param name: Name of Jenkins view, ``str``
        :returns: view configuration (XML format)
        """
        folder_url, short_name = self.getJobFolder(name)
        request = requests.Request("GET", self.buildUrl(CONFIG_VIEW, locals()))
        return self.jenkinsOpen(request)

    def getPromotionName(self, name, job_name):
        """Return the name of a promotion using the API.

        That is roughly an identity method which can be used to
        quickly verify a promotion exists for a job or is accessible
        without causing too much stress on the server side.

        :param name: Promotion name, ``str``
        :param job_name: Job name, ``str``
        :returns: Name of promotion or None
        """
        folder_url, short_name = self.getJobFolder(job_name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(PROMOTION_NAME, locals())))
        except NotFoundException:
            return None
        else:
            actual = json.loads(response)["name"]
            if actual != name:
                raise JenkinsException(
                    "Jenkins returned an unexpected promotion name %s " "(expected: %s)" % (actual, name),
                )
            return actual

    def assertPromotionExists(
        self,
        name,
        job_name,
        exception_message="promotion[%s] does not " "exist for job[%s]",
    ):
        """Raise an exception if a job lacks a promotion

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        :param exception_message: Message to use for the exception. Formatted
                                  with ``name`` and ``job_name``
        :throws: :class:`JenkinsException` whenever the promotion
            does not exist on a job
        """
        if not self.promotionExists(name, job_name):
            raise JenkinsException(exception_message % (name, job_name))

    def promotionExists(self, name, job_name):
        """Check whether a job has a certain promotion

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        :returns: ``True`` if Jenkins promotion exists
        """
        return self.getPromotionName(name, job_name) == name

    def getPromotionsInfo(self, job_name, depth=0):
        """Get promotion information dictionary of a job

        :param job_name: job_name, ``str``
        :param depth: JSON depth, ``int``
        :returns: Dictionary of promotion info, ``dict``
        """
        folder_url, short_name = self.getJobFolder(job_name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(PROMOTION_INFO, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException("job[%s] does not exist" % job_name)
        except req_exc.HTTPError:
            raise JenkinsException("job[%s] does not exist" % job_name)
        except ValueError:
            raise JenkinsException("Could not parse JSON info for " "promotions of job[%s]" % job_name)

    def getPromotions(self, job_name):
        """Get list of promotions running.

        Each promotion is a dictionary with 'name' and 'url' keys.

        :param job_name: Job name, ``str``
        :returns: list of promotions, ``[ { str: str} ]``
        """
        return self.getPromotionsInfo(job_name)["processes"]

    def deletePromotion(self, name, job_name):
        """Delete Jenkins promotion permanently.

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        """
        folder_url, short_name = self.getJobFolder(job_name)
        self.jenkinsOpen(requests.Request("POST", self.buildUrl(DELETE_PROMOTION, locals())))
        if self.promotionExists(name, job_name):
            raise JenkinsException("delete[%s] from job[%s] failed" % (name, job_name))

    def createPromotion(self, name, job_name, config_xml):
        """Create a new Jenkins promotion

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        :param config_xml: config file text, ``str``
        """
        if self.promotionExists(name, job_name):
            raise JenkinsException("promotion[%s] already exists at job[%s]" % (name, job_name))

        folder_url, short_name = self.getJobFolder(job_name)
        self.jenkinsOpen(
            requests.Request(
                "POST",
                self.buildUrl(CREATE_PROMOTION, locals()),
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )
        self.assertPromotionExists(name, job_name, "create[%s] at " "job[%s] failed")

    def reconfigPromotion(self, name, job_name, config_xml):
        """Change configuration of existing Jenkins promotion.

        To create a new promotion, see :meth:`Jenkins.createPromotion`.

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        :param config_xml: New XML configuration, ``str``
        """
        folder_url, short_name = self.getJobFolder(job_name)
        reconfig_url = self.buildUrl(CONFIG_PROMOTION, locals())
        self.jenkinsOpen(
            requests.Request(
                "POST",
                reconfig_url,
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )

    def getPromotionConfig(self, name, job_name):
        """Get configuration of existing Jenkins promotion.

        :param name: Name of Jenkins promotion, ``str``
        :param job_name: Job name, ``str``
        :returns: promotion configuration (XML format)
        """
        folder_url, short_name = self.getJobFolder(job_name)
        request = requests.Request("GET", self.buildUrl(CONFIG_PROMOTION, locals()))
        return self.jenkinsOpen(request)

    def getTagText(self, name, xml):
        """Get text of tag from xml

        :param name: XML tag name, ``str``
        :param xml: XML configuration, ``str``
        :returns: Text of tag, ``str``
        :throws: :class:`JenkinsException` whenever tag does not exist
            or has invalidated text
        """
        tag = ET.fromstring(xml).find(name)
        try:
            text = tag.text.strip()
            if text:
                return text
            raise JenkinsException("tag[%s] is invalidated" % name)
        except AttributeError:
            raise JenkinsException("tag[%s] is invalidated" % name)

    def assertFolder(self, name, exception_message="job[%s] is not a folder"):
        """Raise an exception if job is not Cloudbees Folder

        :param name: Name of job, ``str``
        :param exception_message: Message to use for the exception.
        :throws: :class:`JenkinsException` whenever the job is
            not Cloudbees Folder
        """
        if not self.isFolder(name):
            raise JenkinsException(exception_message % name)

    def isFolder(self, name):
        """Check whether a job is Cloudbees Folder

        :param name: Job name, ``str``
        :returns: ``True`` if job is folder, ``False`` otherwise
        """
        return self.getJobInfo(name)["_class"] == "com.cloudbees.hudson.plugins.folder.Folder"

    def assertCredentialExists(
        self,
        name,
        folder_name,
        domain_name="_",
        exception_message="credential[%s] does not " "exist in the domain[%s] of [%s]",
    ):
        """Raise an exception if credential does not exist in domain of folder

        :param name: Name of credential, ``str``
        :param folder_name: Folder name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        :param exception_message: Message to use for the exception.
                                  Formatted with ``name``, ``domain_name``,
                                  and ``folder_name``
        :throws: :class:`JenkinsException` whenever the credentail
            does not exist in domain of folder
        """
        if not self.credentialExists(name, folder_name, domain_name):
            raise JenkinsException(exception_message % (name, domain_name, folder_name))

    def credentialExists(self, name, folder_name, domain_name="_"):
        """Check whether a credentail exists in domain of folder

        :param name: Name of credentail, ``str``
        :param folder_name: Folder name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        :returns: ``True`` if credentail exists, ``False`` otherwise
        """
        try:
            return self.getCredentialInfo(name, folder_name, domain_name)["id"] == name
        except JenkinsException:
            return False

    def getCredentialInfo(self, name, folder_name, domain_name="_"):
        """Get credential information dictionary in domain of folder

        :param name: Name of credentail, ``str``
        :param folder_name: folder_name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        :returns: Dictionary of credential info, ``dict``
        """
        self.assertFolder(folder_name)
        folder_url, short_name = self.getJobFolder(folder_name)
        try:
            response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(CREDENTIAL_INFO, locals())))
            if response:
                return json.loads(response)
            else:
                raise JenkinsException(
                    "credential[%s] does not exist " "in the domain[%s] of [%s]" % (name, domain_name, folder_name),
                )
        except (req_exc.HTTPError, NotFoundException):
            raise JenkinsException(
                "credential[%s] does not exist " "in the domain[%s] of [%s]" % (name, domain_name, folder_name),
            )
        except ValueError:
            raise JenkinsException(
                "Could not parse JSON info for credential[%s] "
                "in the domain[%s] of [%s]" % (name, domain_name, folder_name),
            )

    def getCredentialConfig(self, name, folder_name, domain_name="_"):
        """Get configuration of credential in domain of folder.

        :param name: Name of credentail, ``str``
        :param folder_name: Folder name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        :returns: Credential configuration (XML format)
        """
        self.assertFolder(folder_name)
        folder_url, short_name = self.getJobFolder(folder_name)
        return self.jenkinsOpen(requests.Request("GET", self.buildUrl(CONFIG_CREDENTIAL, locals())))

    def createCredential(self, folder_name, config_xml, domain_name="_"):
        """Create credentail in domain of folder

        :param folder_name: Folder name, ``str``
        :param config_xml: New XML configuration, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        """
        folder_url, short_name = self.getJobFolder(folder_name)
        name = self.getTagText("id", config_xml)
        if self.credentialExists(name, folder_name, domain_name):
            raise JenkinsException(
                "credential[%s] already exists " "in the domain[%s] of [%s]" % (name, domain_name, folder_name),
            )

        self.jenkinsOpen(
            requests.Request(
                "POST",
                self.buildUrl(CREATE_CREDENTIAL, locals()),
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )
        self.assertCredentialExists(
            name,
            folder_name,
            domain_name,
            "create[%s] failed in the " "domain[%s] of [%s]",
        )

    def deleteCredential(self, name, folder_name, domain_name="_"):
        """Delete credential from domain of folder

        :param name: Name of credentail, ``str``
        :param folder_name: Folder name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        """
        folder_url, short_name = self.getJobFolder(folder_name)
        self.jenkinsOpen(requests.Request("DELETE", self.buildUrl(CONFIG_CREDENTIAL, locals())))
        if self.credentialExists(name, folder_name, domain_name):
            raise JenkinsException(
                "delete credential[%s] from " "domain[%s] of [%s] failed" % (name, domain_name, folder_name),
            )

    def reconfigCredential(self, folder_name, config_xml, domain_name="_"):
        """Reconfig credential with new config in domain of folder

        :param folder_name: Folder name, ``str``
        :param config_xml: New XML configuration, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        """
        folder_url, short_name = self.getJobFolder(folder_name)
        name = self.getTagText("id", config_xml)
        self.assertCredentialExists(name, folder_name, domain_name)

        reconfig_url = self.buildUrl(CONFIG_CREDENTIAL, locals())

        self.jenkinsOpen(
            requests.Request(
                "POST",
                reconfig_url,
                data=config_xml.encode("utf-8"),
                headers=DEFAULT_HEADERS,
            ),
        )

    def listCredentials(self, folder_name, domain_name="_"):
        """List credentials in domain of folder

        :param folder_name: Folder name, ``str``
        :param domain_name: Domain name, default is '_', ``str``
        :returns: Credentials list, ``list``
        """
        self.assertFolder(folder_name)
        folder_url, short_name = self.getJobFolder(folder_name)
        response = self.jenkinsOpen(requests.Request("GET", self.buildUrl(LIST_CREDENTIALS, locals())))
        return json.loads(response)["credentials"]

    def quietDown(self):
        """Prepare Jenkins for shutdown.

        No new builds will be started allowing running builds to complete
        prior to shutdown of the server.
        """
        request = requests.Request("POST", self.buildUrl(QUIET_DOWN))
        self.jenkinsOpen(request)
        info = self.getInfo()
        if not info["quietingDown"]:
            raise JenkinsException("quiet down failed")

    def waitForNormalOp(self, timeout):
        """Wait for jenkins to enter normal operation mode.

        :param timeout: number of seconds to wait, ``int``
            Note this is not the same as the connection timeout set via
            __init__ as that controls the socket timeout. Instead this is
            how long to wait until the status returned.
        :returns: ``True`` if Jenkins became ready in time, ``False``
                   otherwise.

        Setting timeout to be less than the configured connection timeout
        may result in this waiting for at least the connection timeout
        length of time before returning. It is recommended that the timeout
        here should be at least as long as any set connection timeout.
        """
        if timeout < 0:
            raise ValueError("Timeout must be >= 0 not %d" % timeout)

        if not self.timeoutWarningIssued and self.timeout != socket._GLOBAL_DEFAULT_TIMEOUT and timeout < self.timeout:
            warnings.warn(
                "Requested timeout to wait for jenkins to resume "
                "normal operations is less than configured "
                "connection timeout. Unexpected behaviour may "
                "occur.",
            )
            self.timeoutWarningIssued = True

        start_time = time.time()

        def isReady():
            # only call getVersion until it returns without exception
            while True:
                if self.getVersion():
                    while True:
                        # json API will only return valid info once Jenkins
                        # is ready, so just check any known field exists
                        # when not in normal mode, most requests will
                        # be ignored or fail
                        yield "mode" in self.getInfo()
                else:
                    yield False

        while True:
            try:
                if next(isReady()):
                    return True
            except (KeyError, JenkinsException):
                # key missing from JSON, empty response or errors in
                # getInfo due to incomplete HTTP responses
                pass
            # check time passed as the communication will also
            # take time
            if time.time() > start_time + timeout:
                break
            time.sleep(1)

        return False


if __name__ == "__main__":
    jk = Jenkins(url=r"http://localhost:8080/jenkins", username="kamalyes", password="QINg0201$")
    print(jk.getAllJobs())
    print(jk.getBuildInfo(name="YamlInterfaceTests", number=9))



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  depends.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from typing import Any, Callable, Coroutine, Optional

from pydantic import conint
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

from custard.limiter import Limiter
from custard.limiter.enums import GlobalVarEnum


class RateLimiter:
    def __init__(
        self,
        times: conint(ge=0) = 1,
        milliseconds: conint(ge=-1) = 0,
        seconds: conint(ge=-1) = 0,
        minutes: conint(ge=-1) = 0,
        hours: conint(ge=-1) = 0,
        identifier: Optional[Callable] = None,
        callback: Optional[Callable] = None,
    ):
        self.times = times
        self.milliseconds = milliseconds + 1000 * seconds + 60000 * minutes + 3600000 * hours
        self.identifier = identifier
        self.callback = callback

    async def _check(self, key):
        redis = Limiter.redis
        pexpire = await redis.evalsha(Limiter.lua_sha, 1, key, str(self.times), str(self.milliseconds))
        return pexpire

    async def __call__(self, request: Request, response: Response):
        if not Limiter.redis:
            raise Exception(GlobalVarEnum.INIT_ERR_MSG)
        route_index = 0
        dep_index = 0
        for i, route in enumerate(request.app.routes):
            if route.path == request.scope["path"] and request.method in route.methods:
                route_index = i
                for j, dependency in enumerate(route.dependencies):
                    if self is dependency.dependency:
                        dep_index = j
                        break

        # moved here because constructor run before app startup
        identifier = self.identifier or Limiter.identifier
        callback = self.callback or Limiter.http_callback
        rate_key = await identifier(request)
        key = f"{Limiter.prefix}:{rate_key}:{route_index}:{dep_index}"
        pexpire = await self._check(key)
        if pexpire != 0:
            return await callback(request, response, pexpire)
        return None


class WebSocketRateLimiter(RateLimiter):
    async def __call__(self, ws: WebSocket, context_key="") -> Coroutine[Any, Any, Any]:
        if not Limiter.redis:
            raise Exception(GlobalVarEnum.INIT_ERR_MSG)
        identifier = self.identifier or Limiter.identifier
        rate_key = await identifier(ws)
        key = f"{Limiter.prefix}:ws:{rate_key}:{context_key}"
        pexpire = await self._check(key)
        callback = self.callback or Limiter.ws_callback
        if pexpire != 0:
            return await callback(ws, pexpire)
        return None



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  hitfilter.py
@Time    :  2022/5/20 2:52 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  过滤敏感词
"""
import os


class DFAFilter:
    """
    Filter Messages from keywords
    Use DFA to keep algorithm perform constantly
    >>> f = DFAFilter()
    >>> f.add("sexy")
    >>> f.filter("hello sexy baby")
    """

    def __init__(self):
        self.keyword_path = [f"{os.path.dirname(os.path.realpath(__file__))}/keywords"]
        self.keyword_chains = {}
        self.delimit = "\x00"

    def add(self, keyword):
        if not isinstance(keyword, str):
            keyword = keyword.decode("utf-8")
        keyword = keyword.lower()
        chars = keyword.strip()
        if not chars:
            return
        level = self.keyword_chains
        for i in range(len(chars)):
            if chars[i] in level:
                level = level[chars[i]]
            else:
                if not isinstance(level, dict):
                    break
                for j in range(i, len(chars)):
                    level[chars[j]] = {}
                    last_level, last_char = level, chars[j]
                    level = level[chars[j]]
                last_level[last_char] = {self.delimit: 0}
                break
            if i == len(chars) - 1:
                level[self.delimit] = 0

    def parse(self, path=None):
        if path is not None:
            self.keyword_path.append(path)
        if isinstance(self.keyword_path, list):
            for index in self.keyword_path:
                with open(index, "r", encoding="utf-8") as file:
                    for keyword in file:
                        self.add(keyword.strip())
            return None
        else:
            return TypeError("文件路径不正确")

    def filter(self, message, repl="*"):
        if not isinstance(message, str):
            message = message.decode("utf-8")
        message = message.lower()
        ret = []
        start = 0
        while start < len(message):
            level = self.keyword_chains
            step_ins = 0
            for char in message[start:]:
                if char in level:
                    step_ins += 1
                    if self.delimit not in level[char]:
                        level = level[char]
                    else:
                        ret.append(repl * step_ins)
                        start += step_ins - 1
                        break
                else:
                    ret.append(message[start])
                    break
            else:
                ret.append(message[start])
            start += 1

        return "".join(ret)

    def is_contain_sensitive_key_word(self, message):
        repl = "_-__-"
        dest_string = self.filter(message=message, repl=repl)
        if repl in dest_string:
            return True
        return False



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  processor.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  数据处理器
"""
import base64
import difflib
import hashlib
import json
import logging
import operator
import re
import time
import xml.dom.minidom
import xml.etree.ElementTree
from collections import Counter
from functools import reduce
from itertools import zip_longest
from typing import Any, Dict, Tuple
from urllib.parse import unquote

from dicttoxml import dicttoxml
from six import binary_type, text_type

from custard.core.xml2dict import Xml2Dict

logger = logging.getLogger(__name__)


SINGLE_UPLOAD_LENGTH = 5 * 1024 * 1024 * 1024  # 单次上传文件最大为5GB
DEFAULT_CHUNK_SIZE = 1024 * 1024  # 计算MD5值时,文件单次读取的块大小为1MB


class DataKitHelper:
    @classmethod
    def sub_dict_value(cls, *args: Tuple[Dict, ...]) -> Dict:
        """
        Adds two or more dicts together. Common keys will have their values added.

        Returns:
        Example:
            >>> t1 = {'a':1, 'b':2}
            >>> t2 = {'b':1, 'c':3}
            >>> t3 = {'d':5}
            >>> DataKit.sub_dict_value(t1, t2, t3)
            {'a': 1, 'c': 3, 'b': 3, 'd': 5}
        """

        counters = [Counter(arg) for arg in args]
        return dict(reduce(operator.add, counters))

    @classmethod
    def get_target_value(cls, dict_map: dict, separate: str = "$.", result: dict = None):
        if result is None:
            result = {}
        """
        递归获取所有的TargetValue
        :param dict_map: 初始data dict类型
        :param separate: 临时节点 str类型
        :param result:  用于存储所有遍历出来的结果 list集合
        :return: {xx,xx,xx} 以字典形式追加
        Example::
            >>> DataKit.get_target_value(dict_map={"TEST_001": "TEST_VALUE001","TEST_002": [{"TEST_VALUE002-001": "VALUE"}, {"TEST_VALUE002-002": "VALUE"}]})
        """
        if isinstance(dict_map, dict):
            for key, value in dict_map.items():
                temp = separate + key + "."
                # 若类型为list 后面还有一维或二位数组类型数据递归找
                if isinstance(value, list):
                    for i in range(len(value)):
                        cls.get_target_value(dict_map=value[i], separate=temp + str(i) + ".")
                # 若类型还是dict,继续遍历
                elif isinstance(value, dict):
                    cls.get_target_value(dict_map=value, separate=temp)
                # str或者int类型时就基本上判定为具体的xxx值
                elif str(value).isdigit():
                    result.update({separate + key: int(value)})
                elif isinstance(value, str):
                    result.update({separate + key: value})
            return result
        else:
            raise TypeError("传入的参数不是dict类型 %s" % (type(dict_map)))

    @classmethod
    def convert_type(cls, dict_map: dict, disable_data: list = []) -> dict:
        """
        将只有数字的键值给强转类型为int
        :param dict_map: 初始data dict类型
        :param disable: 不用处理的键值对
        Example::
            >>> DataKit.convert_type({'product': {'brand_id': None, 'category_id': '15888'}})
            >>> DataKit.convert_type({'product': {'brand_id': None, 'category_id': '15888'}},["category_id"])
        """
        if isinstance(dict_map, dict):
            for key in list(dict_map.keys()):
                if isinstance(dict_map[key], list):
                    for i in range(len(dict_map[key])):
                        dict_map[key][i] = cls.convert_type(dict_map=dict_map[key][i], disable_data=disable_data)
                elif isinstance(dict_map[key], dict):
                    dict_map[key] = cls.convert_type(dict_map=dict_map[key], disable_data=disable_data)
                elif str(dict_map[key]).isdigit() and str(key) not in disable_data:
                    dict_map[key] = int(dict_map[key])
                elif str(dict_map[key]) == "null":  # 统一处理str无法转化None
                    dict_map[key] = None
            return dict_map
        else:
            raise TypeError("传入的参数不是dict类型 %s" % (type(dict_map)))

    @classmethod
    def dict_to_from(cls, post_data):
        """
        字典转xwww-from格式
        :param post_data: dict {"a": 1, "b":2}
        :return: str: a=1&b=2
        """
        if isinstance(post_data, dict):
            return "&".join(["{}={}".format(key, value) for key, value in post_data.items()])
        return post_data

    @classmethod
    def form_to_dict(cls, post_data):
        """
        x-www-from格式转字典
        :param post_data (str): a=1&b=2
        :return dict: {"a":1, "b":2}
        """
        if isinstance(post_data, str):
            converted_dict = {}
            for k_v in post_data.split("&"):
                try:
                    key, value = k_v.split("=")
                except ValueError:
                    err = "Invalid x_www_form_urlencoded data format: {}".format(post_data)
                    raise Exception(err)
                converted_dict[key] = unquote(value)
            return converted_dict
        else:
            return post_data

    @classmethod
    def json_to_schema(cls, data, result=None):
        """
        json递归生成schema
        :param data:
        :param result:
        :return:
        Example::
            >>> data = {"code": 200, "message": "Success", "error": "",
            ... "ShopInfoList": [{"shop_id": "ML0057", "shop_name": "" }]}
            >>> DataKit.json_to_schema(data=data)
        """
        if result is None:
            result = []
        if isinstance(data, dict):
            is_null = True
            result.append("{")
            result.append("'type': 'object',")
            result.append("'properties': {")
            for k, v in data.items():
                is_null = False
                result.append("'%s':" % k)
                cls.json_to_schema(v, result)
                result.append(",")
            if not is_null:
                result.pop()
            result.append("}")
            result.append("}")
        elif isinstance(data, list):
            result.append("{")
            result.append("'type': 'array',")
            result.append("'items': ")
            cls.json_to_schema(data[0], result)
            result.append("}")
        elif isinstance(data, int):
            result.append("{")
            result.append("'type': 'number'")
            result.append("}")
        elif isinstance(data, str):
            result.append("{")
            result.append("'type': 'string'")
            result.append("}")
        return json.dumps("".join(result), indent=4)

    @classmethod
    def args_to_str(*args, **kwargs):
        """
        args 转换为 str
        Args:
            *args:
            **kwargs:
        Returns:
        """
        str1 = ", ".join(str(i) for i in args)
        kv = []
        for k, v in kwargs.items():
            kv.append(f"{k}={v}")
        str2 = ", ".join(kv)
        if kwargs and args:
            return f"{str1}, {str2}"
        if args:
            return str1
        if kwargs:
            return str2
        return ""

    @classmethod
    def to_str(cls, variable):
        """非字符串转换为字符串"""
        if isinstance(variable, (text_type, binary_type)):
            return variable
        return str(variable)

    @classmethod
    def to_unicode(cls, variable):
        """将字符串转为unicode"""
        if isinstance(variable, binary_type):
            try:
                return variable.decode("utf-8")
            except UnicodeDecodeError:
                raise UnicodeDecodeError("your bytes strings can not be decoded in utf8, utf8 support only!")
        return variable

    @classmethod
    def to_bytes(cls, variable):
        """将字符串转为bytes"""
        if isinstance(variable, text_type):
            try:
                return variable.encode("utf-8")
            except UnicodeEncodeError:
                raise UnicodeEncodeError("your unicode strings can not encoded in utf8, utf8 support only!")
        return variable

    @classmethod
    def deep_dict_update(cls, main_dict: Dict[Any, Any], update_dict: Dict[Any, Any]) -> None:
        """
        字典更新
        Args:
            main_dict:
            update_dict:

        Returns:

        Examples:
            >>> main_dict = {"a":1, "b":2, "c":{"abc":123}}
            >>> update_dict = {"a":2, "b":5, "c":{"abc":1235678}}
            >>> print(DataKitHelper.deep_dict_update(main_dict, update_dict))

        """
        for key, value in update_dict.items():
            if key in main_dict and isinstance(main_dict[key], dict) and isinstance(value, dict):
                cls.deep_dict_update(main_dict[key], value)
            else:
                main_dict[key] = value
        return main_dict

    @classmethod
    def lower_dict_keys(cls, origin_dict: Dict):
        """
        convert keys in dict to lower case
        Args:
            origin_dict: mapping data structure

        Returns:
            dict: mapping with all keys lowered.

        Examples:
            >>> origin_dict = { "Name": "", "Request": "", "URL": "", "METHOD": "", "Headers": "", "Data": ""}
            >>> DataKitHelper.lower_dict_keys(origin_dict)
        """
        if not origin_dict or not isinstance(origin_dict, dict):
            return origin_dict

        return {str(key).lower(): value for key, value in origin_dict.items()}

    @classmethod
    def upper_dict_keys(cls, origin_dict: Dict):
        """
        convert keys in dict to lower case
        Args:
            origin_dict: mapping data structure

        Returns:
            dict: mapping with all keys lowered.

        Examples:
            >>> origin_dict = {'name': '', 'request': '', 'url': '', 'method': '', 'headers': '', 'data': ''}
            >>> DataKitHelper.upper_dict_keys(origin_dict)
        """
        if not origin_dict or not isinstance(origin_dict, dict):
            return origin_dict

        return {str(key).upper(): value for key, value in origin_dict.items()}

    @classmethod
    def omit_long_data(cls, body, omit_len=512):
        """omit too long str/bytes"""
        if not isinstance(body, (str, bytes)):
            return body

        body_len = len(body)
        if body_len <= omit_len:
            return body

        omitted_body = body[0:omit_len]

        appendix_str = f" ... OMITTED {body_len - omit_len} CHARACTORS ..."
        if isinstance(body, bytes):
            appendix_str = appendix_str.encode("utf-8")

        return omitted_body + appendix_str

    @classmethod
    def get_raw_md5(cls, data):
        """计算md5 md5的输入必须为bytes"""
        m2 = hashlib.md5(DataKitHelper.to_bytes(data))
        etag = '"' + str(m2.hexdigest()) + '"'
        return etag

    @classmethod
    def get_md5(cls, data):
        """
        计算 base64 md5 md5的输入必须为bytes
        Args:
            data:

        Returns:

        """
        m2 = hashlib.md5(cls.to_bytes(data))
        md5 = base64.standard_b64encode(m2.digest())
        return md5

    @classmethod
    def get_content_md5(cls, body):
        """计算任何输入流的md5值"""
        if isinstance(body, (text_type, binary_type)):
            return cls.get_md5(body)
        elif hasattr(body, "tell") and hasattr(body, "seek") and hasattr(body, "read"):
            file_position = body.tell()  # 记录文件当前位置
            # avoid OOM
            md5 = hashlib.md5()
            chunk = body.read(DEFAULT_CHUNK_SIZE)
            while chunk:
                md5.update(cls.to_bytes(chunk))
                chunk = body.read(DEFAULT_CHUNK_SIZE)
            md5_str = base64.standard_b64encode(md5.digest())
            try:
                body.seek(file_position)  # 恢复初始的文件位置
            except Exception:
                raise Exception("seek unsupported to calculate md5!")
            return md5_str
        else:
            raise Exception("unsupported body type to calculate md5!")

    @classmethod
    def dict_to_xml(cls, data):
        """V5使用xml格式,将输入的dict转换为xml"""
        doc = xml.dom.minidom.Document()
        root = doc.createElement("CompleteMultipartUpload")
        doc.appendChild(root)

        if "Part" not in data:
            raise Exception("Invalid Parameter, Part Is Required!")

        for i in data["Part"]:
            node_part = doc.createElement("Part")

            if "PartNumber" not in i:
                raise Exception("Invalid Parameter, PartNumber Is Required!")

            node_number = doc.createElement("PartNumber")
            node_number.appendChild(doc.createTextNode(str(i["PartNumber"])))

            if "ETag" not in i:
                raise Exception("Invalid Parameter, ETag Is Required!")

            node_etag = doc.createElement("ETag")
            node_etag.appendChild(doc.createTextNode(str(i["ETag"])))

            node_part.appendChild(node_number)
            node_part.appendChild(node_etag)
            root.appendChild(node_part)
        return doc.toxml("utf-8")

    @classmethod
    def xml_to_dict(cls, data, origin_str="", replace_str=""):
        """V5使用xml格式,将response中的xml转换为dict"""
        root = xml.etree.ElementTree.fromstring(data)
        xmldict = Xml2Dict(root)
        xmlstr = str(xmldict)
        if origin_str:
            xmlstr = xmlstr.replace(origin_str, replace_str)
        xmldict = eval(xmlstr)
        return xmldict

    @classmethod
    def get_id_from_xml(cls, data, name):
        """解析xml中的特定字段"""
        tree = xml.dom.minidom.parseString(data)
        root = tree.documentElement
        result = root.getElementsByTagName(name)
        # use childNodes to get a list, if has no child get itself
        return result[0].childNodes[0].nodeValue

    @classmethod
    def format_xml(cls, data, root, lst=[], parent_child=False):
        """将dict转换为xml, xml_config是一个bytes"""
        if parent_child:
            xml_config = dicttoxml(data, item_func=lambda x: x[:-1], custom_root=root, attr_type=False)
        else:
            xml_config = dicttoxml(data, item_func=lambda x: x, custom_root=root, attr_type=False)
        for i in lst:
            xml_config = xml_config.replace(cls.to_bytes(i + i), cls.to_bytes(i))
        return xml_config

    @classmethod
    def format_values(cls, data):
        """格式化headers和params中的values为bytes"""
        for i in data:
            data[i] = cls.to_bytes(data[i])
        return data

    @classmethod
    def is_json(cls, raw_data, encoding="utf-8"):
        """
        判断目标源是否为json类型
        Args:
            target_data:
        Returns:
        Examples:
            >>> list = ["1235678",{"key1":"value", "key2":"value"}]
            >>> DataKitHelper.is_json()
        """
        if isinstance(raw_data, str):
            try:
                json.loads(raw_data, encoding=encoding)
            except ValueError:
                return False
            return True
        else:
            return False

    @classmethod
    def duplicate(cls, iterable, keep=lambda x: x, key=lambda x: x, reverse=False):
        """
        保序去重
        Args:
            iterable: 去重的同时要对element做的操作
            keep: 使用哪一部分去重
            key: 是否反向去重
            reverse:
        Returns:
        Examples:
            >>> repetition_list = [3, 4, 5, 2, 4, 1]
            # 正序去重
            >>> print(DataKitHelper.duplicate(repetition_list))
            # 逆序去重
            >>> print(DataKitHelper.duplicate(repetition_list, reverse=True))
            # 指定规则去重
            >>> repetition_list = [{"a": 3, "b": 4}, {"a":3, "b": 5}]
            >>> print(DataKitHelper.duplicate(repetition_list, key=lambda x: x["a"]))
            # 去重后仅保留部分数据
            >>> print(DataKitHelper.duplicate(repetition_list, key=lambda x: x["a"], keep=lambda x: x["b"]))
        """
        result = []
        duplicator = []
        if reverse:
            iterable = reversed(iterable)
        for i in iterable:
            keep_field = keep(i)
            key_words = key(i)
            if key_words not in duplicator:
                result.append(keep_field)
                duplicator.append(key_words)
        return list(reversed(result)) if reverse else result

    @classmethod
    def chain_all(cls, iter):
        """
        连接多个序列或字典
        Args:
            iter:
        Returns:
        Examples:
            >>> print(DataKitHelper.chain_all([[1, 2], [1, 2]]))
            >>> print(DataKitHelper.chain_all([{"a": 1}, {"b": 2}]))
        """
        iter = list(iter)
        if not iter:
            return []
        if isinstance(iter[0], dict):
            result = {}
            for i in iter:
                result.update(i)
        else:
            result = reduce(lambda x, y: list(x) + list(y), iter)
        return result

    @classmethod
    def safely_json_loads(
        cls,
        value,
        object_hook: dict = None,
        parse_float=None,
        parse_int=None,
        parse_constant=None,
        object_pairs_hook=None,
        err_detail="解析JSON字符串并将其转换为Python字典失败",
    ):
        """
        返回安全的json类型
        Args:
            value (_type_): _description_
            object_hook (_type_, optional): _description_. Defaults to None.
            parse_float (_type_, optional): _description_. Defaults to None.
            parse_int (_type_, optional): _description_. Defaults to None.
            parse_constant (_type_, optional): _description_. Defaults to None.
            object_pairs_hook (_type_, optional): _description_. Defaults to None.
            err_detail (str, optional): _description_. Defaults to "解析JSON字符串并将其转换为Python字典失败".

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_
        """
        try:
            value = json.loads(
                value,
                object_hook=object_hook,
                parse_float=parse_float,
                parse_int=parse_int,
                parse_constant=parse_constant,
                object_pairs_hook=object_pairs_hook,
            )
        except Exception as e:
            raise Exception(f"{err_detail}: {e}")
        return value

    @classmethod
    def safely_json_dumps(
        cls,
        obj,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        indent=None,
        separators=None,
        default=None,
        sort_keys=False,
        err_detail="解析obj序列化为JSON格式字符串失败",
    ):
        try:
            value = json.dumps(
                obj,
                skipkeys=skipkeys,
                ensure_ascii=ensure_ascii,
                check_circular=check_circular,
                allow_nan=allow_nan,
                indent=indent,
                separators=separators,
                default=default,
                sort_keys=sort_keys,
            )
        except Exception as e:
            raise Exception(f"{err_detail}: {e}")
        return value

    @classmethod
    def format_html_string(cls, html):
        """
        格式化html, 去掉多余的字符,类,script等。
        Args:
            html:
        Returns:
        """
        trims = [
            (r"\n", ""),
            (r"\t", ""),
            (r"\r", ""),
            (r"  ", ""),
            (r"\u2018", "'"),
            (r"\u2019", "'"),
            (r"\ufeff", ""),
            (r"\u2022", ":"),
            (r"<([a-z][a-z0-9]*)\ [^>]*>", "<\\g<1>>"),
            (r"<\s*script[^>]*>[^<]*<\s*/\s*script\s*>", ""),
            (r"</?a.*?>", ""),
        ]
        return reduce(
            lambda string, replacement: re.sub(replacement[0], replacement[1], string),
            trims,
            html,
        )

    @classmethod
    def data_type_convert(cls, original, target_type):
        """
        数据类型转化
        Args:
            original:
            target_type:
        Returns:
        Examples:
            # 将两个相同长度的列表转换成字典
            >>> print(DataKitHelper.data_type_convert(original=(["key1","key2"],["value1","value2"]), target_type="dict"))
            # 将两个不同长度的列表转换成字典
            >>> print(DataKitHelper.data_type_convert(original=(["key1","key2","key3"],["value1","value2"]), target_type="dict"))
            # 将json转化为字典
            >>> print(DataKitHelper.data_type_convert(original='{"errcode": 401,"errs": "[POST]","data": null}', target_type="dict"))
            # 将dict转化为json
            >>> print(DataKitHelper.data_type_convert(original={'errcode': 401,'errs': 'POST','data': True}, target_type="json"))
            # 将字典列表转换为单个字典
            >>> print(DataKitHelper.data_type_convert(original=[{"errcode": 401},{"errs": "[POST]","data": True}], target_type="dict"))
        """
        if isinstance(original, dict) and target_type == "json":
            return json.dumps(original)
        elif isinstance(original, tuple) and target_type == "dict":
            is_equal_bool = len(original[0]) == len(original[1])
            if is_equal_bool:
                return dict(original)
            else:
                return dict(zip_longest(original[0], original[1]))
        elif isinstance(original, list) and target_type == "dict":
            temp = {}
            for index in original:
                temp.update(index)
            return temp
        elif cls.is_json(original) and target_type == "dict":
            return cls.safely_json_loads(original)
        return None

    @classmethod
    def parser(cls, keyword):
        """
        获取后续的字符对应前面出现过的字符的下标
        Args:
            keyword:
        Returns:
        Examples:
            >>> print(DataKitHelper.parser(keyword="55588ABCEACE"))
        """
        vals_mapping = {}
        indices_mapping = {}
        for index, i in enumerate(keyword):
            if i in vals_mapping:
                indices_mapping[index] = vals_mapping[i]
            else:
                indices_mapping[index] = None
                vals_mapping[i] = index
        return indices_mapping

    @classmethod
    def charged(cls, pay_num: int, money_num: list):
        """
        计算余额
        Args:
            pay_num:
            money_num:
        Returns:
        Examples:
            >>> print(DataKitHelper.charged(13, [1, 3, 5]))
        """
        if pay_num == 0:
            return 0
        return min(DataKitHelper.get_min(pay_num, money_num)) + 1

    @classmethod
    def get_min(cls, pay_num, money_num):
        """
        Args:
            pay_num:
            money_num:
        Returns:
        """
        for money in money_num:
            last_num = pay_num - money
            if last_num < 0:
                continue
            try:
                yield cls.charged(last_num, money_num)
            except ValueError:
                continue


class Snowflake:
    epoch = 1292978355588
    worker_id_bits = 5
    data_center_id_bits = 5
    max_worker_id = -1 ^ (-1 << worker_id_bits)
    max_data_center_id = -1 ^ (-1 << data_center_id_bits)
    sequence_bits = 12
    worker_id_shift = sequence_bits
    data_center_id_shift = sequence_bits + worker_id_bits
    timestamp_left_shift = sequence_bits + worker_id_bits + data_center_id_bits
    sequence_mask = -1 ^ (-1 << sequence_bits)

    @classmethod
    def snowflake_to_timestamp(cls, _id):
        _id = _id >> 22  # strip the lower 22 bits
        _id += cls.epoch  # adjust for twitter epoch
        _id = _id / 1000  # convert from milliseconds to seconds
        return _id

    @classmethod
    def generator(cls, worker_id, data_center_id, sleep=lambda x: time.sleep(x / 1000.0)):
        """
        生成器
        Args:
            worker_id (_type_): _description_
            data_center_id (_type_): _description_
            sleep (_type_, optional): _description_. Defaults to lambda x:time.sleep(x / 1000.0).
        Examples:
            >>> s = generator(1, 1)
            >>> for _i in range(1000000):
            ...     print(s.__next__())
        Yields:
            _type_: _description_
        """
        assert 0 <= worker_id <= cls.max_worker_id
        assert 0 <= data_center_id <= cls.max_data_center_id

        last_timestamp = -1
        sequence = 0

        while True:
            timestamp = int(time.time() * 1000)

            if last_timestamp > timestamp:
                print("clock is moving backwards. waiting until %i" % last_timestamp)
                sleep(last_timestamp - timestamp)
                continue

            if last_timestamp == timestamp:
                sequence = (sequence + 1) & cls.sequence_mask
                if sequence == 0:
                    print("sequence overrun")
                    sequence = -1 & cls.sequence_mask
                    sleep(1)
                    continue
            else:
                sequence = 0

            last_timestamp = timestamp
            yield (
                ((timestamp - cls.epoch) << cls.timestamp_left_shift)
                | (data_center_id << cls.data_center_id_shift)
                | (worker_id << cls.worker_id_shift)
                | sequence
            )



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from math import ceil
from typing import Callable, Union

from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

from ..limiter.enums import GlobalVarEnum
from ..limiter.execres import RateLimitException


async def default_identifier(request: Union[Request, WebSocket]):
    forwarded = request.headers.get("X-Forwarded-For")
    ip = forwarded.split(",")[0] if forwarded else request.client.host
    return ip + ":" + request.scope["path"]


async def http_default_callback(request: Request, response: Response, pexpire: int) -> RateLimitException:
    """
    default callback when too many requests
    :param request:
    :param pexpire: The remaining milliseconds
    :param response:
    :return:
    """
    expire = ceil(pexpire / 1000)
    raise RateLimitException(headers={"Retry-After": str(expire)})


async def ws_default_callback(ws: WebSocket, pexpire: int) -> RateLimitException:
    """
    default callback when too many requests
    :param ws:
    :param pexpire: The remaining milliseconds
    :return:
    """
    expire = ceil(pexpire / 1000)
    raise RateLimitException(headers={"Retry-After": str(expire)})


class Limiter:
    redis = None
    prefix: str = None
    lua_sha: str = None
    identifier: Callable = None
    http_callback: Callable = None
    ws_callback: Callable = None
    lua_script = """local key = KEYS[1]
local limit = tonumber(ARGV[1])
local expire_time = ARGV[2]
local current = tonumber(redis.call('get', key) or "0")
if current > 0 then
 if current + 1 > limit then
 return redis.call("PTTL",key)
 else
        redis.call("INCR", key)
 return 0
 end
else
    redis.call("SET", key, 1,"px",expire_time)
 return 0
end"""

    @classmethod
    async def init(
        cls,
        redis,
        prefix: str = f"{GlobalVarEnum.APP_NAME.lower()}:limiter",
        identifier: Callable = default_identifier,
        http_callback: Callable = http_default_callback,
        ws_callback: Callable = ws_default_callback,
    ):
        cls.redis = redis
        cls.prefix = prefix
        cls.identifier = identifier
        cls.http_callback = http_callback
        cls.ws_callback = ws_callback
        cls.lua_sha = await redis.script_load(cls.lua_script)

    @classmethod
    async def close(cls):
        await cls.redis.close()



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  captcha.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  This library is provided to allow standard python logging to output log data as JSON formatted strings
"""
import base64
import os
import random
import time
from io import BytesIO
from string import ascii_letters, digits

from PIL import Image, ImageDraw, ImageFilter, ImageFont

ascii_letters = (
    ascii_letters.replace("i", "").replace("l", "").replace("I", "").replace("Q", "").replace("j", "").replace("q", "")
)
digits = digits.replace("0", "")
BASE_DIR = os.path.dirname(os.path.realpath(__file__))
FONT = [
    BASE_DIR + "/font1.ttf",
    BASE_DIR + "/font2.ttf",
    BASE_DIR + "/font3.ttf",
    BASE_DIR + "/font4.ttf",
]  # https://fonts.google.com/


class CaptchaPainter:
    def __fill_color(self):
        """
        获取颜色
        :return:
        """
        if self.mode == "RGB" or self.mode == "RGBW":
            return random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), random.randint(150, 255)
        else:
            return random.randint(0, 155)

    def __add_chips(self, _draw):
        """
        添加碎片
        :param _draw:
        :return:
        """
        for _chip in range(self.gran * 4):
            # 碎片宽度
            _width = random.randint(self.chip_w - 10, self.chip_w + 10)
            # 碎片高度
            _height = random.randint(self.chip_h - 10, self.chip_h + 10)
            # 碎片左上角 x 轴位置
            _x = random.randint(0, self.im_x)
            # 碎片左上角 y 轴位置
            _y = random.randint(0, self.im_y)
            if _chip % 3 == 1:
                _x2 = random.randint(_x - 50, _x + 50)
                _y2 = random.randint(_y - 50, _y + 50)
                _draw.line((_x, _y, _x2, _y2), fill=self.__fill_color(), width=random.randint(4, 8))
            elif _chip % 3 == 2:
                _draw.ellipse((_x, _y, _x + _width, _y + _height), fill=self.__fill_color())
            else:
                _draw.rectangle((_x, _y, _x + _width, _y + _height), fill=self.__fill_color())

    def __add_text(self):
        """
        添加文字
        :return:
        """
        text_len = len(self.text)
        for _ in range(len(self.text)):
            _x = random.randint(self.im_x // text_len * _, self.im_x // (text_len + 1) * (_ + 1))
            _y = random.randint(self.im_y // 15, self.im_y // 7)
            # print(_x, _y)
            self.draw.text(
                (_x, _y),
                text=self.text[_],
                fill=self.__fill_color(),
                font=ImageFont.truetype(
                    font=random.choice(self.font),
                    size=random.randint(self.im_y - (self.im_y // 4), self.im_y - (self.im_y // 16)),
                ),
            )

    def scribble(self):
        for _ in range(self.gran * 10):
            _x = random.randint(0, self.im_x)
            _y = random.randint(0, self.im_y)
            self.draw.rectangle((_x, _y, _x + 5, _y + 5), fill=self.bg)

    @property
    def normal(self) -> Image:
        """
        静态图片
        :return:
        """
        self.__add_chips(self.draw)
        return self.im

    def gif(self) -> BytesIO:
        """
        动态图片
        :return:
        """
        pics = []
        for _ in range(5):
            _im = self.im.copy()
            _d = ImageDraw.Draw(_im)
            self.__add_chips(_d)
            pics.append(_im)
        buffer = BytesIO()
        pics[0].save(
            buffer,
            save_all=True,
            format="gif",
            append_images=pics[1:],
            transparency=0,
            duration=500,
            loop=0,
            disposal=2,
            quality=9,
        )
        return buffer

    def __init__(self, text="", im_x=260, im_y=125, gran=5, mode="RGB", font: list = None, bg="white"):
        # 验证码文字
        if font is None:
            font = [BASE_DIR + "/font1.ttf"]
        self.text = text
        self.font = font
        # 背景颜色
        self.bg = bg
        # 验证码宽度
        self.im_x = im_x
        # 验证码高度
        self.im_y = im_y
        # 混淆粒度
        self.gran = gran
        # 色彩模式
        self.mode = mode
        # 创建图片
        self.im = Image.new(self.mode, (im_x, im_y), self.bg)
        self.draw = ImageDraw.Draw(self.im)
        # 混淆平均宽度
        self.chip_w = self.im_x // 15
        # 混淆平均高度
        self.chip_h = self.im_y // 10
        # 添加文字
        self.__add_text()
        # self.scribble()


class Captcha:
    @staticmethod
    def __rand(string, k):
        return "".join(random.choices(string, k=k))

    def __filter(self, _im):
        if self.contour:
            return _im.filter(ImageFilter.CONTOUR)
        elif self.enhance:
            return _im.filter(ImageFilter.EDGE_ENHANCE_MORE)
        elif self.edge:
            return _im.filter(ImageFilter.FIND_EDGES)
        elif self.emboss:
            return _im.filter(ImageFilter.EMBOSS)
        else:
            return _im

    def __init__(
        self,
        width=200,
        height=80,
        chips=5,
        mode="RGB",
        imageObj=False,
        gif=False,
        font: list = None,
        bg="white",
        contour=False,
        enhance=False,
        edge=False,
        emboss=False,
    ):
        """

        :param width: 验证码宽度
        :param height: 验证码高度
        :param chips: 混淆碎片数量
        :param mode: 颜色模式("RGB" | "L")
        :param imageObj: 返回 Image 对象
        :param gif: 返回 gif 图片(默认关闭,降低大小,开启时无法使用滤镜)
        :param font: List 字体
        :param bg: 背景颜色(例如:#000000、white)
        :param contour: 轮廓滤镜(ImageFilter.CONTOUR)
        :param enhance: 增强滤镜(ImageFilter.EDGE_ENHANCE_MORE)
        :param edge: 边缘滤镜(ImageFilter.FIND_EDGES) 黑色背景
        :param emboss: 浮雕滤镜(ImageFilter.EMBOSS)
        """
        self.width = width
        self.height = height
        self.chips = chips
        self.mode = mode
        self.font = FONT or font
        self.bg = bg
        self.imageObj = imageObj
        self.gif = gif

        self.contour = contour
        self.enhance = enhance
        self.edge = edge
        self.emboss = emboss

    def letter(self, length=4):
        """
        纯英文
        :param length: 字符长度
        :return:
        """
        text = self.__rand(ascii_letters, length)
        return self.__handel(text)

    def digit(self, length=4):
        """
        纯数字
        :param length: 字符长度
        :return:
        """
        text = self.__rand(digits, length)
        return self.__handel(text)

    def letter_digit(self, length=4):
        """
        数英混合
        :param length:
        :return:
        """
        text = self.__rand(ascii_letters + digits, length)
        return self.__handel(text)

    def __handel(self, text):
        cp = CaptchaPainter(
            text=text,
            im_x=self.width,
            im_y=self.height,
            gran=self.chips,
            mode=self.mode,
            font=self.font,
            bg=self.bg,
        )
        if self.gif:
            data = cp.gif()
            if not self.imageObj:
                return text, "data:image/gif;base64," + base64.b64encode(data.getvalue()).decode()
            else:
                # gif 无法载入为动图
                return text, Image.open(data)
        else:
            im = self.__filter(cp.normal)
            if not self.imageObj:
                buffer = BytesIO()
                im.save(buffer, format="JPEG")
                return text, "data:image/jpeg;base64," + base64.b64encode(buffer.getvalue()).decode()
            else:
                return text, im


if __name__ == "__main__":
    a = Captcha(imageObj=True)

    def test1():
        start = time.time()
        for _ in range(100):
            a.letter_digit()
        print(f"100次测试平均单张速度:{(time.time() - start) / 100}s")

    def test2():
        credit = 0
        while True:
            t, i = a.letter_digit()
            i.show()
            if t.lower() != input(f"<{credit}>请输入验证码:\t").lower():
                break
            credit += 1

    def example():
        # 创建一个 宽 260px 高 125px 混淆碎片度 10 灰度图,返回 (str, Image)
        e = Captcha(width=260, height=125, chips=5, mode="L", imageObj=True, gif=False)
        e1_t, e1_c = e.letter_digit(5)
        print(e1_t, e1_c.show())
        # 将色彩模式改为彩色
        e.mode = "RGB"
        e1_t, e1_c = e.letter(4)
        print(e1_t, e1_c.show())
        # 将碎片减少为 5
        e.chips = 5
        e1_t, e1_c = e.digit(4)
        print(e1_t, e1_c.show())
        # 将背景改为黑色,并使用 emboss 滤镜
        e.bg = "#000000"
        e.emboss = True
        e1_t, e1_c = e.letter_digit()
        print(e1_t, e1_c.show())
        # 自行创建
        e2 = CaptchaPainter(text="Test", gran=5)
        e2.normal.show()

    test1()
    # test2()
    example()



from fastapi import status

from custard.limiter.enums import GlobalVarEnum


class RateLimitException(Exception):
    def __init__(
        self,
        headers: dict,
        code: int = 429,
        detail: str = GlobalVarEnum.INIT_ERR_MSG,
        status_code: int = status.HTTP_429_TOO_MANY_REQUESTS,
    ):
        self.code = code
        self.headers = headers
        self.detail = detail
        self.status_code = status_code



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  xprint
@Time    :  2022/6/3 5:05 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import sys
import time

print_raw = print


def stdout_write(msg: str):
    sys.stdout.write(msg)
    sys.stdout.flush()


def stderr_write(msg: str):
    sys.stderr.write(msg)
    sys.stderr.flush()


def reverse_patch_print():
    """
    提供一个反猴子补丁,恢复print原状
    :return:
    """

    try:
        __builtins__.print = print_raw
    except AttributeError:
        __builtins__["print"] = print_raw


def xprint(*args, sep=" ", end="\n", file=None, color=True):
    """
    超流弊的print补丁
    Args:
        *args:
        sep:
        end:
        file:
        color:
    Returns:
    Examples:
        >>> xprint("123")
    """
    args = (str(arg) for arg in args)  # REMIND 防止是数字不能被join
    if file == sys.stderr:
        stderr_write(sep.join(args))  # 如 threading 模块第926行,打印线程错误,希望保持原始的红色错误方式,不希望转成蓝色。
    elif file in [sys.stdout, None]:
        # 获取被调用函数在被调用时所处代码行数
        line = sys._getframe().f_back.f_lineno
        # 获取被调用函数所在模块文件名
        file_name = sys._getframe(1).f_code.co_filename
        single_space = "\t"
        if color:
            stdout_write(
                f'\033[35m{time.strftime("%H:%M:%S")}{single_space}"{file_name}:{line}"{single_space}'
                f"{sep.join(args)}{single_space}{end}\033[0m",
            )
        else:
            stdout_write(f'{time.strftime("%H:%M:%S")}\t"{file_name}:{line}"{single_space}{sep.join(args)}{end}')
    else:
        print_raw(*args, sep=sep, end=end, file=file)
